# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    'load_component',
    'load_component_from_text',
    'load_component_from_url',
    'load_component_from_file',
]

import copy
import sys
from collections import OrderedDict
from typing import Any, List, Mapping, NamedTuple, Sequence, Union
from ._naming import _sanitize_file_name, _sanitize_python_function_name, generate_unique_name_conversion_table
from ._yaml_utils import load_yaml
from .structures import *
from ._data_passing import serialize_value, get_canonical_type_for_type_struct


_default_component_name = 'Component'


def load_component(filename=None, url=None, text=None):
    '''
    Loads component from text, file or URL and creates a task factory function
    
    Only one argument should be specified.

    Args:
        filename: Path of local file containing the component definition.
        url: The URL of the component file data
        text: A string containing the component file data.

    Returns:
        A factory function with a strongly-typed signature.
        Once called with the required arguments, the factory constructs a pipeline task instance (ContainerOp).
    '''
    #This function should be called load_task_factory since it returns a factory function.
    #The real load_component function should produce an object with component properties (e.g. name, description, inputs/outputs).
    #TODO: Change this function to return component spec object but it should be callable to construct tasks.
    non_null_args_count = len([name for name, value in locals().items() if value != None])
    if non_null_args_count != 1:
        raise ValueError('Need to specify exactly one source')
    if filename:
        return load_component_from_file(filename)
    elif url:
        return load_component_from_url(url)
    elif text:
        return load_component_from_text(text)
    else:
        raise ValueError('Need to specify a source')


def load_component_from_url(url):
    '''
    Loads component from URL and creates a task factory function
    
    Args:
        url: The URL of the component file data

    Returns:
        A factory function with a strongly-typed signature.
        Once called with the required arguments, the factory constructs a pipeline task instance (ContainerOp).
    '''
    component_spec = _load_component_spec_from_url(url)
    url = _fix_component_uri(url)
    component_ref = ComponentReference(url=url)
    return _create_task_factory_from_component_spec(
        component_spec=component_spec,
        component_filename=url,
        component_ref=component_ref,
    )


def load_component_from_file(filename):
    '''
    Loads component from file and creates a task factory function
    
    Args:
        filename: Path of local file containing the component definition.

    Returns:
        A factory function with a strongly-typed signature.
        Once called with the required arguments, the factory constructs a pipeline task instance (ContainerOp).
    '''
    component_spec = _load_component_spec_from_file(path=filename)
    return _create_task_factory_from_component_spec(
        component_spec=component_spec,
        component_filename=filename,
    )


def load_component_from_text(text):
    '''
    Loads component from text and creates a task factory function
    
    Args:
        text: A string containing the component file data.

    Returns:
        A factory function with a strongly-typed signature.
        Once called with the required arguments, the factory constructs a pipeline task instance (ContainerOp).
    '''
    if text is None:
        raise TypeError
    component_spec = _load_component_spec_from_component_text(text)
    return _create_task_factory_from_component_spec(component_spec=component_spec)


def _fix_component_uri(uri: str) -> str:
    #Handling Google Cloud Storage URIs
    if uri.startswith('gs://'):
        #Replacing the gs:// URI with https:// URI (works for public objects)
        uri = 'https://storage.googleapis.com/' + uri[len('gs://'):]
    return uri


def _load_component_spec_from_file(path) -> ComponentSpec:
    with open(path, 'rb') as component_stream:
        return _load_component_spec_from_yaml_or_zip_bytes(component_stream.read())


def _load_component_spec_from_url(url: str):
    if url is None:
        raise TypeError

    url = _fix_component_uri(url)

    import requests
    resp = requests.get(url)
    resp.raise_for_status()
    return _load_component_spec_from_yaml_or_zip_bytes(resp.content)


_COMPONENT_FILE_NAME_IN_ARCHIVE = 'component.yaml'


def _load_component_spec_from_yaml_or_zip_bytes(data: bytes):
    '''Loads component spec from binary data.

    The data can be a YAML file or a zip file with a component.yaml file inside.
    '''
    import zipfile
    import io
    stream = io.BytesIO(data)
    if zipfile.is_zipfile(stream):
        stream.seek(0)
        with zipfile.ZipFile(stream) as zip_obj:
            data = zip_obj.read(_COMPONENT_FILE_NAME_IN_ARCHIVE)
    return _load_component_spec_from_component_text(data)


def _load_component_spec_from_component_text(text) -> ComponentSpec:
    component_dict = load_yaml(text)
    component_spec = ComponentSpec.from_dict(component_dict)

    # Calculating hash digest for the component
    import hashlib
    data = text if isinstance(text, bytes) else text.encode('utf-8')
    data = data.replace(b'\r\n', b'\n')  # Normalizing line endings
    digest = hashlib.sha256(data).hexdigest()
    component_spec._digest = digest

    return component_spec


_inputs_dir = '/tmp/inputs'
_outputs_dir = '/tmp/outputs'
_single_io_file_name = 'data'


def _generate_input_file_name(port_name):
    return _inputs_dir + '/' + _sanitize_file_name(port_name) + '/' + _single_io_file_name


def _generate_output_file_name(port_name):
    return _outputs_dir + '/' + _sanitize_file_name(port_name) + '/' + _single_io_file_name


def _react_to_incompatible_reference_type(
    input_type,
    argument_type,
    input_name: str,
):
    """Raises error for the case when the argument type is incompatible with the input type."""
    message = 'Argument with type "{}" was passed to the input "{}" that has type "{}".'.format(argument_type, input_name, input_type)
    raise TypeError(message)


def _create_task_spec_from_component_and_arguments(
    component_spec: ComponentSpec,
    arguments: Mapping[str, Any],
    component_ref: ComponentReference = None,
) -> TaskSpec:
    """Constructs a TaskSpec object from component reference and arguments.
    The function also checks the arguments types and serializes them."""
    if component_ref is None:
        component_ref = ComponentReference(spec=component_spec)
    else:
        component_ref = copy.copy(component_ref)
        component_ref.spec = component_spec

    # Not checking for missing or extra arguments since the dynamic factory function checks that
    task_arguments = {}
    for input_name, argument_value in arguments.items():
        input_type = component_spec._inputs_dict[input_name].type

        if isinstance(argument_value, (GraphInputArgument, TaskOutputArgument)):
            # argument_value is a reference
            if isinstance(argument_value, GraphInputArgument):
                reference_type = argument_value.graph_input.type
            elif isinstance(argument_value, TaskOutputArgument):
                reference_type = argument_value.task_output.type
            else:
                reference_type = None

            if reference_type and input_type and reference_type != input_type:
                _react_to_incompatible_reference_type(input_type, reference_type, input_name)

            task_arguments[input_name] = argument_value
        else:
            # argument_value is a constant value
            serialized_argument_value = serialize_value(argument_value, input_type)
            task_arguments[input_name] = serialized_argument_value

    task = TaskSpec(
        component_ref=component_ref,
        arguments=task_arguments,
    )
    task._init_outputs()

    return task


_default_container_task_constructor = _create_task_spec_from_component_and_arguments

# Holds the function that constructs a task object based on ComponentSpec, arguments and ComponentReference.
# Framework authors can override this constructor function to construct different framework-specific task-like objects.
# The task object should have the task.outputs dictionary with keys corresponding to the ComponentSpec outputs.
# The default constructor creates and instance of the TaskSpec class.
_container_task_constructor = _default_container_task_constructor


_always_expand_graph_components = False


class _DefaultValue:
    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return repr(self.value)


#TODO: Refactor the function to make it shorter
def _create_task_factory_from_component_spec(component_spec:ComponentSpec, component_filename=None, component_ref: ComponentReference = None):
    name = component_spec.name or _default_component_name

    func_docstring_lines = []
    if component_spec.name:
        func_docstring_lines.append(component_spec.name)
    if component_spec.description:
        func_docstring_lines.append(component_spec.description)
    
    inputs_list = component_spec.inputs or [] #List[InputSpec]
    input_names = [input.name for input in inputs_list]

    #Creating the name translation tables : Original <-> Pythonic 
    input_name_to_pythonic = generate_unique_name_conversion_table(input_names, _sanitize_python_function_name)
    pythonic_name_to_input_name = {v: k for k, v in input_name_to_pythonic.items()}

    if component_ref is None:
        component_ref = ComponentReference(spec=component_spec, url=component_filename)
    else:
        component_ref.spec = component_spec
    
    digest = getattr(component_spec, '_digest', None)
    # TODO: Calculate the digest if missing
    if digest:
        # TODO: Report possible digest conflicts
        component_ref.digest = digest


    def create_task_from_component_and_arguments(pythonic_arguments):
        arguments = {
            pythonic_name_to_input_name[argument_name]: argument_value
            for argument_name, argument_value in pythonic_arguments.items()
            if not isinstance(argument_value, _DefaultValue) # Skipping passing arguments for optional values that have not been overridden.
        }

        if (
            isinstance(component_spec.implementation, GraphImplementation)
            and (
                # When the container task constructor is not overriden, we just construct TaskSpec for both container and graph tasks.
                # If the container task constructor is overriden, we should expand the graph components so that the override is called for all sub-tasks.
                _container_task_constructor != _default_container_task_constructor
                or _always_expand_graph_components
            )
        ):
            return _resolve_graph_task(
                component_spec=component_spec,
                arguments=arguments,
                component_ref=component_ref,
            )

        task = _container_task_constructor(
            component_spec=component_spec,
            arguments=arguments,
            component_ref=component_ref,
        )

        return task

    import inspect
    from . import _dynamic

    #Reordering the inputs since in Python optional parameters must come after required parameters
    reordered_input_list = [input for input in inputs_list if input.default is None and not input.optional] + [input for input in inputs_list if not (input.default is None and not input.optional)]

    def component_default_to_func_default(component_default: str, is_optional: bool):
        if is_optional:
            return _DefaultValue(component_default)
        if component_default is not None:
            return component_default
        return inspect.Parameter.empty

    input_parameters = [
        _dynamic.KwParameter(
            input_name_to_pythonic[port.name],
            annotation=(get_canonical_type_for_type_struct(str(port.type)) or str(port.type) if port.type else inspect.Parameter.empty),
            default=component_default_to_func_default(port.default, port.optional),
        )
        for port in reordered_input_list
    ]
    factory_function_parameters = input_parameters #Outputs are no longer part of the task factory function signature. The paths are always generated by the system.
    
    task_factory = _dynamic.create_function_from_parameters(
        create_task_from_component_and_arguments,        
        factory_function_parameters,
        documentation='\n'.join(func_docstring_lines),
        func_name=name,
        func_filename=component_filename if (component_filename and (component_filename.endswith('.yaml') or component_filename.endswith('.yml'))) else None,
    )
    task_factory.component_spec = component_spec
    return task_factory


_ResolvedCommandLineAndPaths = NamedTuple(
    '_ResolvedCommandLineAndPaths',
    [
        ('command', Sequence[str]),
        ('args', Sequence[str]),
        ('input_paths', Mapping[str, str]),
        ('output_paths', Mapping[str, str]),
        ('inputs_consumed_by_value', Mapping[str, str]),
    ],
)


def _resolve_command_line_and_paths(
    component_spec: ComponentSpec,
    arguments: Mapping[str, str],
    input_path_generator=_generate_input_file_name,
    output_path_generator=_generate_output_file_name,
    argument_serializer=serialize_value,
) -> _ResolvedCommandLineAndPaths:
    """Resolves the command line argument placeholders. Also produces the maps of the generated inpuit/output paths."""
    argument_values = arguments

    if not isinstance(component_spec.implementation, ContainerImplementation):
        raise TypeError('Only container components have command line to resolve')

    inputs_dict = {input_spec.name: input_spec for input_spec in component_spec.inputs or []}
    container_spec = component_spec.implementation.container

    output_paths = OrderedDict() #Preserving the order to make the kubernetes output names deterministic
    unconfigurable_output_paths = container_spec.file_outputs or {}
    for output in component_spec.outputs or []:
        if output.name in unconfigurable_output_paths:
            output_paths[output.name] = unconfigurable_output_paths[output.name]

    input_paths = OrderedDict()
    inputs_consumed_by_value = {}

    def expand_command_part(arg) -> Union[str, List[str], None]:
        if arg is None:
            return None
        if isinstance(arg, (str, int, float, bool)):
            return str(arg)

        if isinstance(arg, InputValuePlaceholder):
            input_name = arg.input_name
            input_spec = inputs_dict[input_name]
            input_value = argument_values.get(input_name, None)
            if input_value is not None:
                serialized_argument = argument_serializer(input_value, input_spec.type)
                inputs_consumed_by_value[input_name] = serialized_argument
                return serialized_argument
            else:
                if input_spec.optional:
                    return None
                else:
                    raise ValueError('No value provided for input {}'.format(input_name))

        if isinstance(arg, InputPathPlaceholder):
            input_name = arg.input_name
            input_value = argument_values.get(input_name, None)
            if input_value is not None:
                input_path = input_path_generator(input_name)
                input_paths[input_name] = input_path
                return input_path
            else:
                input_spec = inputs_dict[input_name]
                if input_spec.optional:
                    #Even when we support default values there is no need to check for a default here.
                    #In current execution flow (called by python task factory), the missing argument would be replaced with the default value by python itself.
                    return None
                else:
                    raise ValueError('No value provided for input {}'.format(input_name))

        elif isinstance(arg, OutputPathPlaceholder):
            output_name = arg.output_name
            output_filename = output_path_generator(output_name)
            if arg.output_name in output_paths:
                if output_paths[output_name] != output_filename:
                    raise ValueError('Conflicting output files specified for port {}: {} and {}'.format(output_name, output_paths[output_name], output_filename))
            else:
                output_paths[output_name] = output_filename

            return output_filename

        elif isinstance(arg, ConcatPlaceholder):
            expanded_argument_strings = expand_argument_list(arg.items)
            return ''.join(expanded_argument_strings)

        elif isinstance(arg, IfPlaceholder):
            arg = arg.if_structure
            condition_result = expand_command_part(arg.condition)
            from distutils.util import strtobool
            condition_result_bool = condition_result and strtobool(condition_result) #Python gotcha: bool('False') == True; Need to use strtobool; Also need to handle None and []
            result_node = arg.then_value if condition_result_bool else arg.else_value
            if result_node is None:
                return []
            if isinstance(result_node, list):
                expanded_result = expand_argument_list(result_node)
            else:
                expanded_result = expand_command_part(result_node)
            return expanded_result

        elif isinstance(arg, IsPresentPlaceholder):
            argument_is_present = argument_values.get(arg.input_name, None) is not None
            return str(argument_is_present)
        else:
            raise TypeError('Unrecognized argument type: {}'.format(arg))
    
    def expand_argument_list(argument_list):
        expanded_list = []
        if argument_list is not None:
            for part in argument_list:
                expanded_part = expand_command_part(part)
                if expanded_part is not None:
                    if isinstance(expanded_part, list):
                        expanded_list.extend(expanded_part)
                    else:
                        expanded_list.append(str(expanded_part))
        return expanded_list

    expanded_command = expand_argument_list(container_spec.command)
    expanded_args = expand_argument_list(container_spec.args)

    return _ResolvedCommandLineAndPaths(
        command=expanded_command,
        args=expanded_args,
        input_paths=input_paths,
        output_paths=output_paths,
        inputs_consumed_by_value=inputs_consumed_by_value,
    )


_ResolvedGraphTask = NamedTuple(
    '_ResolvedGraphTask',
    [
        ('component_spec', ComponentSpec),
        ('component_ref', ComponentReference),
        ('outputs', Mapping[str, Any]),
        ('task_arguments', Mapping[str, Any]),
    ],
)


def _resolve_graph_task(
    component_spec: ComponentSpec,
    arguments: Mapping[str, Any],
    component_ref: ComponentReference = None,
) -> TaskSpec:
    from ..components import ComponentStore
    component_store = ComponentStore.default_store

    graph = component_spec.implementation.graph

    graph_input_arguments = {input.name: input.default for input in component_spec.inputs or [] if input.default is not None}
    graph_input_arguments.update(arguments)

    outputs_of_tasks = {}
    def resolve_argument(argument):
        if isinstance(argument, (str, int, float, bool)):
            return argument
        elif isinstance(argument, GraphInputArgument):
            return graph_input_arguments[argument.graph_input.input_name]
        elif isinstance(argument, TaskOutputArgument):
            upstream_task_output_ref = argument.task_output
            upstream_task_outputs = outputs_of_tasks[upstream_task_output_ref.task_id]
            upstream_task_output = upstream_task_outputs[upstream_task_output_ref.output_name]
            return upstream_task_output
        else:
            raise TypeError('Argument for input has unexpected type "{}".'.format(type(argument)))

    for task_id, task_spec in graph._toposorted_tasks.items(): # Cannot use graph.tasks here since they might be listed not in dependency order. Especially on python <3.6 where the dicts do not preserve ordering
        task_factory = component_store._load_component_from_ref(task_spec.component_ref)
        # TODO: Handle the case when optional graph component input is passed to optional task component input
        task_arguments = {input_name: resolve_argument(argument) for input_name, argument in task_spec.arguments.items()}
        task_component_spec = task_factory.component_spec

        input_name_to_pythonic = generate_unique_name_conversion_table([input.name for input in task_component_spec.inputs or []], _sanitize_python_function_name)
        output_name_to_pythonic = generate_unique_name_conversion_table([output.name for output in task_component_spec.outputs or []], _sanitize_python_function_name)
        pythonic_output_name_to_original = {pythonic_name: original_name for original_name, pythonic_name in output_name_to_pythonic.items()}
        pythonic_task_arguments = {input_name_to_pythonic[input_name]: argument for input_name, argument in task_arguments.items()}

        task_obj = task_factory(**pythonic_task_arguments)
        task_outputs_with_pythonic_names = task_obj.outputs
        task_outputs_with_original_names = {pythonic_output_name_to_original[pythonic_output_name]: output_value for pythonic_output_name, output_value in task_outputs_with_pythonic_names.items()}
        outputs_of_tasks[task_id] = task_outputs_with_original_names

    resolved_graph_outputs = OrderedDict([(output_name, resolve_argument(argument)) for output_name, argument in graph.output_values.items()])

    # For resolved graph component tasks task.outputs point to the actual tasks that originally produced the output that is later returned from the graph
    graph_task = _ResolvedGraphTask(
        component_ref=component_ref,
        component_spec=component_spec,
        outputs = resolved_graph_outputs,
        task_arguments=arguments,
    )
    return graph_task



# Lint as: python3
# Copyright 2019 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License,Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for diagnose_me.gcp."""

from typing import Text
import unittest
from unittest import mock
from . import gcp
from . import utility


class GoogleCloudTest(unittest.TestCase):

  @mock.patch.object(gcp, 'execute_gcloud_command', autospec=True)
  def test_project_configuration_gcloud(self, mock_execute_gcloud_command):
    """Tests gcloud commands."""
    gcp.get_gcp_configuration(gcp.Commands.GET_APIS)
    mock_execute_gcloud_command.assert_called_once_with(['services', 'list'],
                                                        project_id=None,
                                                        human_readable=False)

  @mock.patch.object(gcp, 'execute_gsutil_command', autospec=True)
  def test_project_configuration_gsutil(self, mock_execute_gsutil_command):
    """Test Gsutil commands."""
    gcp.get_gcp_configuration(gcp.Commands.GET_STORAGE_BUCKETS)
    mock_execute_gsutil_command.assert_called_once_with(['ls'], project_id=None)

  def test_Commands(self):
    """Verify commands are formaated properly."""
    for command in gcp.Commands:
      self.assertIsInstance(gcp._command_string[command], Text)
      self.assertNotIn('\t', gcp._command_string[command])
      self.assertNotIn('\n', gcp._command_string[command])

  @mock.patch.object(utility, 'ExecutorResponse', autospec=True)
  def test_execute_gsutil_command(self, mock_executor_response):
    """Test execute_gsutil_command."""
    gcp.execute_gsutil_command(
        [gcp._command_string[gcp.Commands.GET_STORAGE_BUCKETS]])
    mock_executor_response().execute_command.assert_called_once_with(
        ['gsutil', 'ls'])

    gcp.execute_gsutil_command(
        [gcp._command_string[gcp.Commands.GET_STORAGE_BUCKETS]],
        project_id='test_project')
    mock_executor_response().execute_command.assert_called_with(
        ['gsutil', 'ls', '-p', 'test_project'])

  @mock.patch.object(utility, 'ExecutorResponse', autospec=True)
  def test_execute_gcloud_command(self, mock_executor_response):
    """Test execute_gcloud_command."""
    gcp.execute_gcloud_command(
        gcp._command_string[gcp.Commands.GET_APIS].split(' '))
    mock_executor_response().execute_command.assert_called_once_with(
        ['gcloud', 'services', 'list', '--format', 'json'])

    gcp.execute_gcloud_command(
        gcp._command_string[gcp.Commands.GET_APIS].split(' '),
        project_id='test_project')
    # verify project id is added correctly
    mock_executor_response().execute_command.assert_called_with([
        'gcloud', 'services', 'list', '--format', 'json', '--project',
        'test_project'
    ])
    # verify human_readable removes json fromat flag
    gcp.execute_gcloud_command(
        gcp._command_string[gcp.Commands.GET_APIS].split(' '),
        project_id='test_project',
        human_readable=True)
    mock_executor_response().execute_command.assert_called_with(
        ['gcloud', 'services', 'list', '--project', 'test_project'])


if __name__ == '__main__':
  unittest.main()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import re
from typing import List, Dict
from kubernetes.client.models import (
    V1ObjectMeta, V1ResourceRequirements, V1PersistentVolumeClaimSpec,
    V1PersistentVolumeClaim, V1TypedLocalObjectReference
)

from ._resource_op import ResourceOp
from ._pipeline_param import (
    PipelineParam, match_serialized_pipelineparam, sanitize_k8s_name
)
from ._pipeline_volume import PipelineVolume


VOLUME_MODE_RWO = ["ReadWriteOnce"]
VOLUME_MODE_RWM = ["ReadWriteMany"]
VOLUME_MODE_ROM = ["ReadOnlyMany"]


class VolumeOp(ResourceOp):
    """Represents an op which will be translated into a resource template
    which will be creating a PVC.
    """

    def __init__(self,
                 resource_name: str = None,
                 size: str = None,
                 storage_class: str = None,
                 modes: List[str] = None,
                 annotations: Dict[str, str] = None,
                 data_source=None,
                 volume_name=None,
                 **kwargs):
        """Create a new instance of VolumeOp.

        Args:
            resource_name: A desired name for the PVC which will be created
            size: The size of the PVC which will be created
            storage_class: The storage class to use for the dynamically created
                PVC
            modes: The access modes for the PVC
            annotations: Annotations to be patched in the PVC
            data_source: May be a V1TypedLocalObjectReference, and then it is
                used in the data_source field of the PVC as is. Can also be a
                string/PipelineParam, and in that case it will be used as a
                VolumeSnapshot name (Alpha feature)
            volume_name: VolumeName is the binding reference to the PersistentVolume
                backing this claim.
            kwargs: See ResourceOp definition
        Raises:
            ValueError: if k8s_resource is provided along with other arguments
                        if k8s_resource is not a V1PersistentVolumeClaim
                        if size is None
                        if size is an invalid memory string (when not a
                            PipelineParam)
                        if data_source is not one of (str, PipelineParam,
                            V1TypedLocalObjectReference)
        """
        # Add size to attribute outputs
        self.attribute_outputs = {"size": "{.status.capacity.storage}"}

        if "k8s_resource" in kwargs:
            if resource_name or size or storage_class or modes or annotations:
                raise ValueError("You cannot provide k8s_resource along with "
                                 "other arguments.")
            if not isinstance(kwargs["k8s_resource"], V1PersistentVolumeClaim):
                raise ValueError("k8s_resource in VolumeOp must be an instance"
                                 " of V1PersistentVolumeClaim")
            super().__init__(**kwargs)
            self.volume = PipelineVolume(
                name=sanitize_k8s_name(self.name),
                pvc=self.outputs["name"]
            )
            return

        if not size:
            raise ValueError("Please provide size")
        elif not match_serialized_pipelineparam(str(size)):
            self._validate_memory_string(size)

        if data_source and not isinstance(
           data_source, (str, PipelineParam, V1TypedLocalObjectReference)):
            raise ValueError("data_source can be one of (str, PipelineParam, "
                             "V1TypedLocalObjectReference).")
        if data_source and isinstance(data_source, (str, PipelineParam)):
            data_source = V1TypedLocalObjectReference(
                api_group="snapshot.storage.k8s.io",
                kind="VolumeSnapshot",
                name=data_source
            )

        # Set the k8s_resource
        if not match_serialized_pipelineparam(str(resource_name)):
            resource_name = sanitize_k8s_name(resource_name)
        pvc_metadata = V1ObjectMeta(
            name="{{workflow.name}}-%s" % resource_name,
            annotations=annotations
        )
        requested_resources = V1ResourceRequirements(
            requests={"storage": size}
        )
        pvc_spec = V1PersistentVolumeClaimSpec(
            access_modes=modes or VOLUME_MODE_RWM,
            resources=requested_resources,
            storage_class_name=storage_class,
            data_source=data_source,
            volume_name=volume_name
        )
        k8s_resource = V1PersistentVolumeClaim(
            api_version="v1",
            kind="PersistentVolumeClaim",
            metadata=pvc_metadata,
            spec=pvc_spec
        )

        super().__init__(
            k8s_resource=k8s_resource,
            **kwargs,
        )
        self.volume = PipelineVolume(
            name=sanitize_k8s_name(self.name),
            pvc=self.outputs["name"]
        )

    def _validate_memory_string(self, memory_string):
        """Validate a given string is valid for memory request or limit."""
        if re.match(r'^[0-9]+(E|Ei|P|Pi|T|Ti|G|Gi|M|Mi|K|Ki){0,1}$',
                    memory_string) is None:
            raise ValueError('Invalid memory string. Should be an integer, ' +
                             'or integer followed by one of ' +
                             '"E|Ei|P|Pi|T|Ti|G|Gi|M|Mi|K|Ki"')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import Callable, Dict, Optional, Text

from ..dsl._container_op import BaseOp, ContainerOp

# Pod label indicating the SDK type from which the pipeline is
# generated. By default it's set to kfp.
_SDK_ENV_LABEL = 'pipelines.kubeflow.org/pipeline-sdk-type'
_SDK_ENV_DEFAULT = 'kfp'

# Common prefix of KFP OOB components url paths.
_OOB_COMPONENT_PATH_PREFIX = 'https://raw.githubusercontent.com/kubeflow/'\
                             'pipelines'

# Key for component origin path pod label.
COMPONENT_PATH_LABEL_KEY = 'pipelines.kubeflow.org/component_origin_path'

# Key for component spec digest pod label.
COMPONENT_DIGEST_LABEL_KEY = 'pipelines.kubeflow.org/component_digest'


def get_default_telemetry_labels() -> Dict[Text, Text]:
    """Returns the default pod labels for telemetry purpose."""
    result = {
        _SDK_ENV_LABEL: _SDK_ENV_DEFAULT,
    }
    return result


def add_pod_env(op: BaseOp) -> BaseOp:
    """Adds pod environment info to ContainerOp.
    """
    if isinstance(op, ContainerOp) and op.pod_labels and op.pod_labels.get('add-pod-env', None) == 'true':
        from kubernetes import client as k8s_client
        op.container.add_env_variable(
            k8s_client.V1EnvVar(
                name='KFP_POD_NAME',
                value_from=k8s_client.V1EnvVarSource(
                    field_ref=k8s_client.V1ObjectFieldSelector(
                        field_path='metadata.name'
                    )
                )
            )
        ).add_env_variable(
            k8s_client.V1EnvVar(
                name='KFP_NAMESPACE',
                value_from=k8s_client.V1EnvVarSource(
                    field_ref=k8s_client.V1ObjectFieldSelector(
                        field_path='metadata.namespace'
                    )
                )
            )
        )
    return op


def add_pod_labels(labels: Optional[Dict[Text, Text]] = None) -> Callable:
    """Adds provided pod labels to each pod."""

    def _add_pod_labels(task):
        for k, v in labels.items():
            # Only append but not update.
            # This is needed to bypass TFX pipelines/components.
            if k not in task.pod_labels:
                task.add_pod_label(k, v)
        return task

    return _add_pod_labels


def _remove_suffix(string: Text, suffix: Text) -> Text:
    """Removes the suffix from a string."""
    if suffix and string.endswith(suffix):
        return string[:-len(suffix)]
    else:
        return string


def add_name_for_oob_components() -> Callable:
    """Adds the OOB component name if applicable."""
    
    def _add_name_for_oob_components(task):
        # Detect the component origin uri in component_ref if exists, and
        # attach the OOB component name as a pod label.
        component_ref = getattr(task, '_component_ref', None)
        if component_ref:
            if component_ref.url:
                origin_path = _remove_suffix(
                    component_ref.url, 'component.yaml').rstrip('/')
                # Only include KFP OOB components.
                if origin_path.startswith(_OOB_COMPONENT_PATH_PREFIX):
                    origin_path = origin_path.split('/', 7)[-1]
                else:
                    return task
                # Clean the label to comply with the k8s label convention.
                origin_path = re.sub('[^-a-z0-9A-Z_.]', '.', origin_path)
                origin_path_label = origin_path[-63:].strip('-_.')
                task.add_pod_label(COMPONENT_PATH_LABEL_KEY, origin_path_label)
            if component_ref.digest:
                task.add_pod_label(
                    COMPONENT_DIGEST_LABEL_KEY, component_ref.digest)
            
        return task
    
    return _add_name_for_oob_components


# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import logging
from tabulate import tabulate


@click.group()
def pipeline():
    """Manage pipeline resources"""
    pass


@pipeline.command()
@click.option(
    "-p",
    "--pipeline-name",
    help="Name of the pipeline."
)
@click.argument("package-file")
@click.pass_context
def upload(ctx, pipeline_name, package_file):
    """Upload a KFP pipeline"""
    client = ctx.obj["client"]
    if not pipeline_name:
        pipeline_name = package_file.split(".")[0]

    pipeline = client.upload_pipeline(package_file, pipeline_name)
    logging.info("Pipeline {} has been submitted\n".format(pipeline.id))
    _display_pipeline(pipeline)


@pipeline.command()
@click.option(
    "-p",
    "--pipeline-id",
    help="ID of the pipeline",
    required=True
)
@click.option(
    "-v",
    "--pipeline-version",
    help="Name of the pipeline version",
    required=True
)
@click.argument("package-file")
@click.pass_context
def upload_version(ctx, package_file, pipeline_version, pipeline_id):
    """Upload a version of the KFP pipeline"""
    client = ctx.obj["client"]

    version = client.pipeline_uploads.upload_pipeline_version(
        package_file, name=pipeline_version, pipelineid=pipeline_id)
    logging.info(
        "The {} version of the pipeline {} has been submitted\n".format(
            pipeline_version, pipeline_id))
    _display_pipeline_version(version)


@pipeline.command()
@click.option(
    "--max-size",
    default=100,
    help="Max size of the listed pipelines."
)
@click.pass_context
def list(ctx, max_size):
    """List uploaded KFP pipelines"""
    client = ctx.obj["client"]

    response = client.list_pipelines(
        page_size=max_size,
        sort_by="created_at desc"
    )
    if response.pipelines:
        _print_pipelines(response.pipelines)
    else:
        logging.info("No pipelines found")


@pipeline.command()
@click.argument("pipeline-id")
@click.pass_context
def get(ctx, pipeline_id):
    """Get detailed information about an uploaded KFP pipeline"""
    client = ctx.obj["client"]

    pipeline = client.get_pipeline(pipeline_id)
    _display_pipeline(pipeline)


@pipeline.command()
@click.argument("pipeline-id")
@click.pass_context
def delete(ctx, pipeline_id):
    """Delete an uploaded KFP pipeline"""
    client = ctx.obj["client"]

    client.delete_pipeline(pipeline_id)
    print("{} is deleted".format(pipeline_id))


def _print_pipelines(pipelines):
    headers = ["Pipeline ID", "Name", "Uploaded at"]
    data = [[
        pipeline.id,
        pipeline.name,
        pipeline.created_at.isoformat()
    ] for pipeline in pipelines]
    print(tabulate(data, headers=headers, tablefmt="grid"))


def _display_pipeline(pipeline):
    # Print pipeline information
    print(tabulate([], headers=["Pipeline Details"]))
    table = [
        ["ID", pipeline.id],
        ["Name", pipeline.name],
        ["Description", pipeline.description],
        ["Uploaded at", pipeline.created_at.isoformat()],
    ]
    print(tabulate(table, tablefmt="plain"))

    # Print pipeline parameter details
    headers = ["Parameter Name", "Default Value"]
    data = [[param.name, param.value] for param in pipeline.parameters]
    print(tabulate(data, headers=headers, tablefmt="grid"))


def _display_pipeline_version(version):
    print(tabulate([], headers=["Pipeline Version Details"]))
    pipeline_id = version.resource_references[0].key.id
    table = [
        ["Pipeline ID", pipeline_id],
        ["Version Name", version.name],
        ["Uploaded at", version.created_at.isoformat()],
    ]
    print(tabulate(table, tablefmt="plain"))



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from collections import namedtuple
from typing import List, Dict, Union


# TODO: Move this to a separate class
# For now, this identifies a condition with only "==" operator supported.
ConditionOperator = namedtuple('ConditionOperator', 'operator operand1 operand2')
PipelineParamTuple = namedtuple('PipelineParamTuple', 'name op pattern')


def sanitize_k8s_name(name, allow_capital_underscore=False):
  """From _make_kubernetes_name
    sanitize_k8s_name cleans and converts the names in the workflow.

  Args:
    name: original name,
    allow_capital_underscore: whether to allow capital letter and underscore
      in this name.

  Returns:
    sanitized name.
  """
  if allow_capital_underscore:
    return re.sub('-+', '-', re.sub('[^-_0-9A-Za-z]+', '-', name)).lstrip('-').rstrip('-')
  else:
    return re.sub('-+', '-', re.sub('[^-0-9a-z]+', '-', name.lower())).lstrip('-').rstrip('-')


def match_serialized_pipelineparam(payload: str):
  """match_serialized_pipelineparam matches the serialized pipelineparam.
  Args:
    payloads (str): a string that contains the serialized pipelineparam.

  Returns:
    PipelineParamTuple
  """
  matches = re.findall(r'{{pipelineparam:op=([\w\s_-]*);name=([\w\s_-]+)}}', payload)
  param_tuples = []
  for match in matches:
      pattern = '{{pipelineparam:op=%s;name=%s}}' % (match[0], match[1])
      param_tuples.append(PipelineParamTuple(
                          name=sanitize_k8s_name(match[1], True),
                          op=sanitize_k8s_name(match[0]), 
                          pattern=pattern))
  return param_tuples


def _extract_pipelineparams(payloads: str or List[str]):
  """_extract_pipelineparam extract a list of PipelineParam instances from the payload string.
  Note: this function removes all duplicate matches.

  Args:
    payload (str or list[str]): a string/a list of strings that contains serialized pipelineparams
  Return:
    List[PipelineParam]
  """
  if isinstance(payloads, str):
    payloads = [payloads]
  param_tuples = []
  for payload in payloads:
    param_tuples += match_serialized_pipelineparam(payload)
  pipeline_params = []
  for param_tuple in list(set(param_tuples)):
    pipeline_params.append(PipelineParam(param_tuple.name,
                                         param_tuple.op, 
                                         pattern=param_tuple.pattern))
  return pipeline_params


def extract_pipelineparams_from_any(payload) -> List['PipelineParam']:
  """Recursively extract PipelineParam instances or serialized string from any object or list of objects.

  Args:
    payload (str or k8_obj or list[str or k8_obj]): a string/a list 
        of strings that contains serialized pipelineparams or a k8 definition 
        object.
  Return:
    List[PipelineParam]
  """
  if not payload:
    return []

  # PipelineParam
  if isinstance(payload, PipelineParam):
    return [payload]
 
  # str
  if isinstance(payload, str):
    return list(set(_extract_pipelineparams(payload)))
  
  # list or tuple
  if isinstance(payload, list) or isinstance(payload, tuple):
    pipeline_params = []
    for item in payload:
      pipeline_params += extract_pipelineparams_from_any(item)
    return list(set(pipeline_params))

  # dict
  if isinstance(payload, dict):
    pipeline_params = []
    for key, value in payload.items():
      pipeline_params += extract_pipelineparams_from_any(key)
      pipeline_params += extract_pipelineparams_from_any(value)
    return list(set(pipeline_params))

  # k8s OpenAPI object
  if hasattr(payload, 'attribute_map') and isinstance(payload.attribute_map, dict):
    pipeline_params = []
    for key in payload.attribute_map:
      pipeline_params += extract_pipelineparams_from_any(getattr(payload, key))

    return list(set(pipeline_params))

  # return empty list  
  return []


class PipelineParam(object):
  """Representing a future value that is passed between pipeline components.

  A PipelineParam object can be used as a pipeline function argument so that it will be a
  pipeline parameter that shows up in ML Pipelines system UI. It can also represent an intermediate
  value passed between components.
  """
  
  def __init__(self, name: str, op_name: str=None, value: str=None, param_type : Union[str, Dict] = None, pattern: str=None):
    """Create a new instance of PipelineParam.
    Args:
      name: name of the pipeline parameter.
      op_name: the name of the operation that produces the PipelineParam. None means
               it is not produced by any operator, so if None, either user constructs it
               directly (for providing an immediate value), or it is a pipeline function
               argument.
      value: The actual value of the PipelineParam. If provided, the PipelineParam is
             "resolved" immediately. For now, we support string only.
      param_type: the type of the PipelineParam.
      pattern: the serialized string regex pattern this pipeline parameter created from. 
    Raises: ValueError in name or op_name contains invalid characters, or both op_name
            and value are set.
    """

    valid_name_regex = r'^[A-Za-z][A-Za-z0-9\s_-]*$'
    if not re.match(valid_name_regex, name):
      raise ValueError('Only letters, numbers, spaces, "_", and "-" are allowed in name. Must begin with a letter.  '
                       'Got name: {}'.format(name))

    if op_name and value:
      raise ValueError('op_name and value cannot be both set.')

    self.name = name
    # ensure value is None even if empty string or empty list
    # so that serialization and unserialization remain consistent
    # (i.e. None => '' => None)
    self.op_name = op_name if op_name else None
    self.value = value if value else None
    self.param_type = param_type
    self.pattern = pattern or str(self)

  @property
  def full_name(self):
    """Unique name in the argo yaml for the PipelineParam"""
    if self.op_name:
        return self.op_name + '-' + self.name
    return self.name

  def __str__(self):
    """String representation.

    The string representation is a string identifier so we can mix the PipelineParam inline
    with other strings such as arguments. For example, we can support:
    ['echo %s' % param] as the container command and later a compiler can replace
    the placeholder "{{pipelineparam:op=%s;name=%s}}" with its own parameter identifier.
    """

    #This is deleted because if users specify default values to PipelineParam,
    # The compiler can not detect it as the value is not NULL.
    #if self.value:
    #  return str(self.value)

    op_name = self.op_name if self.op_name else ''
    return '{{pipelineparam:op=%s;name=%s}}' % (op_name, self.name)
  
  def __repr__(self):
    # return str({self.__class__.__name__: self.__dict__})
    # We make repr return the placeholder string so that if someone uses str()-based serialization of complex objects containing `PipelineParam` it works properly (e.g. str([1, 2, 3, kfp.dsl.PipelineParam("aaa"), 4, 5, 6,]))
    return str(self)

  def to_struct(self):
    # Used by the json serializer. Outputs a JSON-serializable representation of the object
    return str(self)

  def __eq__(self, other):
    return ConditionOperator('==', self, other)

  def __ne__(self, other):
    return ConditionOperator('!=', self, other)

  def __lt__(self, other):
    return ConditionOperator('<', self, other)

  def __le__(self, other):
    return ConditionOperator('<=', self, other)

  def __gt__(self, other):
    return ConditionOperator('>', self, other)

  def __ge__(self, other):
    return ConditionOperator('>=', self, other)

  def __hash__(self):
    return hash((self.op_name, self.name))

  def ignore_type(self):
    """ignore_type ignores the type information such that type checking would also pass"""
    self.param_type = None
    return self




# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    'InputSpec',
    'OutputSpec',

    'InputValuePlaceholder',
    'InputPathPlaceholder',
    'OutputPathPlaceholder',
    'ConcatPlaceholder',
    'IsPresentPlaceholder',
    'IfPlaceholderStructure',
    'IfPlaceholder',

    'ContainerSpec',
    'ContainerImplementation',

    'ComponentSpec',

    'ComponentReference',

    'GraphInputReference',
    'GraphInputArgument',
    'TaskOutputReference',
    'TaskOutputArgument',

    'EqualsPredicate',
    'NotEqualsPredicate',
    'GreaterThanPredicate',
    'GreaterThanOrEqualPredicate',
    'LessThenPredicate',
    'LessThenOrEqualPredicate',
    'NotPredicate',
    'AndPredicate',
    'OrPredicate',

    'RetryStrategySpec',
    'CachingStrategySpec',
    'ExecutionOptionsSpec',
    'TaskSpec',

    'GraphSpec',
    'GraphImplementation',

    'PipelineRunSpec',
]

from collections import OrderedDict

from typing import Any, Dict, List, Mapping, Optional, Sequence, Union

from .modelbase import ModelBase

from .structures.kubernetes import v1


PrimitiveTypes = Union[str, int, float, bool]
PrimitiveTypesIncludingNone = Optional[PrimitiveTypes]

TypeSpecType = Union[str, Dict, List]


class InputSpec(ModelBase):
    '''Describes the component input specification'''
    def __init__(self,
        name: str,
        type: Optional[TypeSpecType] = None,
        description: Optional[str] = None,
        default: Optional[PrimitiveTypes] = None,
        optional: Optional[bool] = False,
    ):
        super().__init__(locals())


class OutputSpec(ModelBase):
    '''Describes the component output specification'''
    def __init__(self,
        name: str,
        type: Optional[TypeSpecType] = None,
        description: Optional[str] = None,
    ):
        super().__init__(locals())


class InputValuePlaceholder(ModelBase): #Non-standard attr names
    '''Represents the command-line argument placeholder that will be replaced at run-time by the input argument value.'''
    _serialized_names = {
        'input_name': 'inputValue',
    }

    def __init__(self,
        input_name: str,
    ):
        super().__init__(locals())


class InputPathPlaceholder(ModelBase): #Non-standard attr names
    '''Represents the command-line argument placeholder that will be replaced at run-time by a local file path pointing to a file containing the input argument value.'''
    _serialized_names = {
        'input_name': 'inputPath',
    }

    def __init__(self,
        input_name: str,
    ):
        super().__init__(locals())


class OutputPathPlaceholder(ModelBase): #Non-standard attr names
    '''Represents the command-line argument placeholder that will be replaced at run-time by a local file path pointing to a file where the program should write its output data.'''
    _serialized_names = {
        'output_name': 'outputPath',
    }

    def __init__(self,
        output_name: str,
    ):
        super().__init__(locals())


CommandlineArgumentType = Union[
    str,
    InputValuePlaceholder,
    InputPathPlaceholder,
    OutputPathPlaceholder,
    'ConcatPlaceholder',
    'IfPlaceholder',
]


class ConcatPlaceholder(ModelBase): #Non-standard attr names
    '''Represents the command-line argument placeholder that will be replaced at run-time by the concatenated values of its items.'''
    _serialized_names = {
        'items': 'concat',
    }

    def __init__(self,
        items: List[CommandlineArgumentType],
    ):
        super().__init__(locals())


class IsPresentPlaceholder(ModelBase): #Non-standard attr names
    '''Represents the command-line argument placeholder that will be replaced at run-time by a boolean value specifying whether the caller has passed an argument for the specified optional input.'''
    _serialized_names = {
        'input_name': 'isPresent',
    }

    def __init__(self,
        input_name: str,
    ):
        super().__init__(locals())


IfConditionArgumentType = Union[bool, str, IsPresentPlaceholder, InputValuePlaceholder]


class IfPlaceholderStructure(ModelBase): #Non-standard attr names
    '''Used in by the IfPlaceholder - the command-line argument placeholder that will be replaced at run-time by the expanded value of either "then_value" or "else_value" depending on the submissio-time resolved value of the "cond" predicate.'''
    _serialized_names = {
        'condition': 'cond',
        'then_value': 'then',
        'else_value': 'else',
    }

    def __init__(self,
        condition: IfConditionArgumentType,
        then_value: Union[CommandlineArgumentType, List[CommandlineArgumentType]],
        else_value: Optional[Union[CommandlineArgumentType, List[CommandlineArgumentType]]] = None,
    ):
        super().__init__(locals())


class IfPlaceholder(ModelBase): #Non-standard attr names
    '''Represents the command-line argument placeholder that will be replaced at run-time by the expanded value of either "then_value" or "else_value" depending on the submissio-time resolved value of the "cond" predicate.'''
    _serialized_names = {
        'if_structure': 'if',
    }

    def __init__(self,
        if_structure: IfPlaceholderStructure,
    ):
        super().__init__(locals())


class ContainerSpec(ModelBase):
    '''Describes the container component implementation.'''
    _serialized_names = {
        'file_outputs': 'fileOutputs', #TODO: rename to something like legacy_unconfigurable_output_paths
    }

    def __init__(self,
        image: str,
        command: Optional[List[CommandlineArgumentType]] = None,
        args: Optional[List[CommandlineArgumentType]] = None,
        env: Optional[Mapping[str, str]] = None,
        file_outputs: Optional[Mapping[str, str]] = None, #TODO: rename to something like legacy_unconfigurable_output_paths
    ):
        super().__init__(locals())


class ContainerImplementation(ModelBase):
    '''Represents the container component implementation.'''
    def __init__(self,
        container: ContainerSpec,
    ):
        super().__init__(locals())


ImplementationType = Union[ContainerImplementation, 'GraphImplementation']


class MetadataSpec(ModelBase):
    def __init__(self,
        annotations: Optional[Dict[str, str]] = None,
        labels: Optional[Dict[str, str]] = None,
    ):
        super().__init__(locals())


class ComponentSpec(ModelBase):
    '''Component specification. Describes the metadata (name, description, annotations and labels), the interface (inputs and outputs) and the implementation of the component.'''
    def __init__(
        self,
        name: Optional[str] = None, #? Move to metadata?
        description: Optional[str] = None, #? Move to metadata?
        metadata: Optional[MetadataSpec] = None,
        inputs: Optional[List[InputSpec]] = None,
        outputs: Optional[List[OutputSpec]] = None,
        implementation: Optional[ImplementationType] = None,
        version: Optional[str] = 'google.com/cloud/pipelines/component/v1',
        #tags: Optional[Set[str]] = None,
    ):
        super().__init__(locals())
        self._post_init()

    def _post_init(self):
        #Checking input names for uniqueness
        self._inputs_dict = {}
        if self.inputs:
            for input in self.inputs:
                if input.name in self._inputs_dict:
                    raise ValueError('Non-unique input name "{}"'.format(input.name))
                self._inputs_dict[input.name] = input

        #Checking output names for uniqueness
        self._outputs_dict = {}
        if self.outputs:
            for output in self.outputs:
                if output.name in self._outputs_dict:
                    raise ValueError('Non-unique output name "{}"'.format(output.name))
                self._outputs_dict[output.name] = output

        if isinstance(self.implementation, ContainerImplementation):
            container = self.implementation.container

            if container.file_outputs:
                for output_name, path in container.file_outputs.items():
                    if output_name not in self._outputs_dict:
                        raise TypeError('Unconfigurable output entry "{}" references non-existing output.'.format({output_name: path}))

            def verify_arg(arg):
                if arg is None:
                    pass
                elif isinstance(arg, (str, int, float, bool)):
                    pass
                elif isinstance(arg, list):
                    for arg2 in arg:
                        verify_arg(arg2)
                elif isinstance(arg, (InputValuePlaceholder, InputPathPlaceholder, IsPresentPlaceholder)):
                    if arg.input_name not in self._inputs_dict:
                        raise TypeError('Argument "{}" references non-existing input.'.format(arg))
                elif isinstance(arg, OutputPathPlaceholder):
                    if arg.output_name not in self._outputs_dict:
                        raise TypeError('Argument "{}" references non-existing output.'.format(arg))
                elif isinstance(arg, ConcatPlaceholder):
                    for arg2 in arg.items:
                        verify_arg(arg2)
                elif isinstance(arg, IfPlaceholder):
                    verify_arg(arg.if_structure.condition)
                    verify_arg(arg.if_structure.then_value)
                    verify_arg(arg.if_structure.else_value)
                else:
                    raise TypeError('Unexpected argument "{}"'.format(arg))
            
            verify_arg(container.command)
            verify_arg(container.args)

        if isinstance(self.implementation, GraphImplementation):
            graph = self.implementation.graph

            if graph.output_values is not None:
                for output_name, argument in graph.output_values.items():
                    if output_name not in self._outputs_dict:
                        raise TypeError('Graph output argument entry "{}" references non-existing output.'.format({output_name: argument}))

            if graph.tasks is not None:
                for task in graph.tasks.values():
                    if task.arguments is not None:
                        for argument in task.arguments.values():
                            if isinstance(argument, GraphInputArgument) and argument.graph_input.input_name not in self._inputs_dict:
                                raise TypeError('Argument "{}" references non-existing input.'.format(argument))

    def save(self, file_path: str):
        '''Saves the component definition to file. It can be shared online and later loaded using the load_component function.'''
        from ._yaml_utils import dump_yaml
        component_yaml = dump_yaml(self.to_dict())
        with open(file_path, 'w') as f:
            f.write(component_yaml)


class ComponentReference(ModelBase):
    '''Component reference. Contains information that can be used to locate and load a component by name, digest or URL'''
    def __init__(self,
        name: Optional[str] = None,
        digest: Optional[str] = None,
        tag: Optional[str] = None,
        url: Optional[str] = None,
        spec: Optional[ComponentSpec] = None,
    ):
        super().__init__(locals())
        self._post_init()
    
    def _post_init(self) -> None:
        if not any([self.name, self.digest, self.tag, self.url, self.spec]):
            raise TypeError('Need at least one argument.')


class GraphInputReference(ModelBase):
    '''References the input of the graph (the scope is a single graph).'''
    _serialized_names = {
        'input_name': 'inputName',
    }

    def __init__(self,
        input_name: str,
        type: Optional[TypeSpecType] = None,    # Can be used to override the reference data type
    ):
        super().__init__(locals())

    def as_argument(self) -> 'GraphInputArgument':
        return GraphInputArgument(graph_input=self)

    def with_type(self, type_spec: TypeSpecType) -> 'GraphInputReference':
        return GraphInputReference(
            input_name=self.input_name,
            type=type_spec,
        )

    def without_type(self) -> 'GraphInputReference':
        return self.with_type(None)

class GraphInputArgument(ModelBase):
    '''Represents the component argument value that comes from the graph component input.'''
    _serialized_names = {
        'graph_input': 'graphInput',
    }

    def __init__(self,
        graph_input: GraphInputReference,
    ):
        super().__init__(locals())


class TaskOutputReference(ModelBase):
    '''References the output of some task (the scope is a single graph).'''
    _serialized_names = {
        'task_id': 'taskId',
        'output_name': 'outputName',
    }

    def __init__(self,
        output_name: str,
        task_id: Optional[str] = None,      # Used for linking to the upstream task in serialized component file.
        task: Optional['TaskSpec'] = None,  # Used for linking to the upstream task in runtime since Task does not have an ID until inserted into a graph.
        type: Optional[TypeSpecType] = None,    # Can be used to override the reference data type
    ):
        super().__init__(locals())
        if self.task_id is None and self.task is None:
            raise TypeError('task_id and task cannot be None at the same time.')

    def with_type(self, type_spec: TypeSpecType) -> 'TaskOutputReference':
        return TaskOutputReference(
            output_name=self.output_name,
            task_id=self.task_id,
            task=self.task,
            type=type_spec,
        )

    def without_type(self) -> 'TaskOutputReference':
        return self.with_type(None)


class TaskOutputArgument(ModelBase): #Has additional constructor for convenience
    '''Represents the component argument value that comes from the output of another task.'''
    _serialized_names = {
        'task_output': 'taskOutput',
    }

    def __init__(self,
        task_output: TaskOutputReference,
    ):
        super().__init__(locals())

    @staticmethod
    def construct(
        task_id: str,
        output_name: str,
    ) -> 'TaskOutputArgument':
        return TaskOutputArgument(TaskOutputReference(
            task_id=task_id,
            output_name=output_name,
        ))

    def with_type(self, type_spec: TypeSpecType) -> 'TaskOutputArgument':
        return TaskOutputArgument(
            task_output=self.task_output.with_type(type_spec),
        )

    def without_type(self) -> 'TaskOutputArgument':
        return self.with_type(None)

ArgumentType = Union[PrimitiveTypes, GraphInputArgument, TaskOutputArgument]


class TwoOperands(ModelBase):
    def __init__(self,
        op1: ArgumentType,
        op2: ArgumentType,
    ):
        super().__init__(locals())


class BinaryPredicate(ModelBase): #abstract base type
    def __init__(self,
        operands: TwoOperands
    ):
        super().__init__(locals())


class EqualsPredicate(BinaryPredicate):
    '''Represents the "equals" comparison predicate.'''
    _serialized_names = {'operands': '=='}


class NotEqualsPredicate(BinaryPredicate):
    '''Represents the "not equals" comparison predicate.'''
    _serialized_names = {'operands': '!='}


class GreaterThanPredicate(BinaryPredicate):
    '''Represents the "greater than" comparison predicate.'''
    _serialized_names = {'operands': '>'}


class GreaterThanOrEqualPredicate(BinaryPredicate):
    '''Represents the "greater than or equal" comparison predicate.'''
    _serialized_names = {'operands': '>='}


class LessThenPredicate(BinaryPredicate):
    '''Represents the "less than" comparison predicate.'''
    _serialized_names = {'operands': '<'}


class LessThenOrEqualPredicate(BinaryPredicate):
    '''Represents the "less than or equal" comparison predicate.'''
    _serialized_names = { 'operands': '<='}


PredicateType = Union[
    ArgumentType,
    EqualsPredicate, NotEqualsPredicate, GreaterThanPredicate, GreaterThanOrEqualPredicate, LessThenPredicate, LessThenOrEqualPredicate,
    'NotPredicate', 'AndPredicate', 'OrPredicate',
]


class TwoBooleanOperands(ModelBase):
    def __init__(self,
        op1: PredicateType,
        op2: PredicateType,
    ):
        super().__init__(locals())


class NotPredicate(ModelBase):
    '''Represents the "not" logical operation.'''
    _serialized_names = {'operand': 'not'}

    def __init__(self,
        operand: PredicateType
    ):
        super().__init__(locals())


class AndPredicate(ModelBase):
    '''Represents the "and" logical operation.'''
    _serialized_names = {'operands': 'and'}

    def __init__(self,
        operands: TwoBooleanOperands
    ) :
        super().__init__(locals())

class OrPredicate(ModelBase):
    '''Represents the "or" logical operation.'''
    _serialized_names = {'operands': 'or'}

    def __init__(self,
        operands: TwoBooleanOperands
    ):
        super().__init__(locals())


class RetryStrategySpec(ModelBase):
    _serialized_names = {
        'max_retries': 'maxRetries',
    }

    def __init__(self,
        max_retries: int,
    ):
        super().__init__(locals())


class CachingStrategySpec(ModelBase):
    _serialized_names = {
        'max_cache_staleness': 'maxCacheStaleness',
    }

    def __init__(self,
        max_cache_staleness: Optional[str] = None,  # RFC3339 compliant duration: P30DT1H22M3S
    ):
        super().__init__(locals())


class KubernetesExecutionOptionsSpec(ModelBase):
    _serialized_names = {
        'main_container': 'mainContainer',
        'pod_spec': 'podSpec',
    }

    def __init__(self,
        metadata: Optional[v1.ObjectMetaArgoSubset] = None,
        main_container: Optional[v1.Container] = None,
        pod_spec: Optional[v1.PodSpecArgoSubset] = None,
    ):
        super().__init__(locals())


class ExecutionOptionsSpec(ModelBase):
    _serialized_names = {
        'retry_strategy': 'retryStrategy',
        'caching_strategy': 'cachingStrategy',
        'kubernetes_options': 'kubernetesOptions',
    }

    def __init__(self,
        retry_strategy: Optional[RetryStrategySpec] = None,
        caching_strategy: Optional[CachingStrategySpec] = None,
        kubernetes_options: Optional[KubernetesExecutionOptionsSpec] = None,
    ):
        super().__init__(locals())


class TaskSpec(ModelBase):
    '''Task specification. Task is a "configured" component - a component supplied with arguments and other applied configuration changes.'''
    _serialized_names = {
        'component_ref': 'componentRef',
        'is_enabled': 'isEnabled',
        'execution_options': 'executionOptions'
    }

    def __init__(self,
        component_ref: ComponentReference,
        arguments: Optional[Mapping[str, ArgumentType]] = None,
        is_enabled: Optional[PredicateType] = None,
        execution_options: Optional[ExecutionOptionsSpec] = None,
    ):
        super().__init__(locals())
        #TODO: If component_ref is resolved to component spec, then check that the arguments correspond to the inputs

    def _init_outputs(self):
        #Adding output references to the task
        if self.component_ref.spec is None:
            return
        task_outputs = OrderedDict()
        for output in self.component_ref.spec.outputs or []:
            task_output_ref = TaskOutputReference(
                output_name=output.name,
                task=self,
                type=output.type, # TODO: Resolve type expressions. E.g. type: {TypeOf: Input 1}
            )
            task_output_arg = TaskOutputArgument(task_output=task_output_ref)
            task_outputs[output.name] = task_output_arg

        self.outputs = task_outputs
        if len(task_outputs) == 1:
            self.output = list(task_outputs.values())[0]


class GraphSpec(ModelBase):
    '''Describes the graph component implementation. It represents a graph of component tasks connected to the upstream sources of data using the argument specifications. It also describes the sources of graph output values.'''
    _serialized_names = {
        'output_values': 'outputValues',
    }

    def __init__(self,
        tasks: Mapping[str, TaskSpec],
        output_values: Mapping[str, ArgumentType] = None,
    ):
        super().__init__(locals())
        self._post_init()
    
    def _post_init(self):
        #Checking task output references and preparing the dependency table
        task_dependencies = {}
        for task_id, task in self.tasks.items():
            dependencies = set()
            task_dependencies[task_id] = dependencies
            if task.arguments is not None:
                for argument in task.arguments.values():
                    if isinstance(argument, TaskOutputArgument):
                        dependencies.add(argument.task_output.task_id)
                        if argument.task_output.task_id not in self.tasks:
                            raise TypeError('Argument "{}" references non-existing task.'.format(argument))

        #Topologically sorting tasks to detect cycles
        task_dependents = {k: set() for k in task_dependencies.keys()}
        for task_id, dependencies in task_dependencies.items():
            for dependency in dependencies:
                task_dependents[dependency].add(task_id)
        task_number_of_remaining_dependencies = {k: len(v) for k, v in task_dependencies.items()}
        sorted_tasks = OrderedDict()
        def process_task(task_id):
            if task_number_of_remaining_dependencies[task_id] == 0 and task_id not in sorted_tasks:
                sorted_tasks[task_id] = self.tasks[task_id]
                for dependent_task in task_dependents[task_id]:
                    task_number_of_remaining_dependencies[dependent_task] = task_number_of_remaining_dependencies[dependent_task] - 1
                    process_task(dependent_task)
        for task_id in task_dependencies.keys():
            process_task(task_id)
        if len(sorted_tasks) != len(task_dependencies):
            tasks_with_unsatisfied_dependencies = {k: v for k, v in task_number_of_remaining_dependencies.items() if v > 0}
            task_wth_minimal_number_of_unsatisfied_dependencies = min(tasks_with_unsatisfied_dependencies.keys(), key=lambda task_id: tasks_with_unsatisfied_dependencies[task_id])
            raise ValueError('Task "{}" has cyclical dependency.'.format(task_wth_minimal_number_of_unsatisfied_dependencies))
        
        self._toposorted_tasks = sorted_tasks


class GraphImplementation(ModelBase):
    '''Represents the graph component implementation.'''
    def __init__(self,
        graph: GraphSpec,
    ):
        super().__init__(locals())


class PipelineRunSpec(ModelBase):
    '''The object that can be sent to the backend to start a new Run.'''
    _serialized_names = {
        'root_task': 'rootTask',
        #'on_exit_task': 'onExitTask',
    }

    def __init__(self,
        root_task: TaskSpec,
        #on_exit_task: Optional[TaskSpec] = None,
    ):
        super().__init__(locals())



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Dict
from kubernetes.client.models import (
    V1Volume, V1TypedLocalObjectReference, V1ObjectMeta
)

from ._resource_op import ResourceOp
from ._pipeline_param import match_serialized_pipelineparam, sanitize_k8s_name


class VolumeSnapshotOp(ResourceOp):
    """Represents an op which will be translated into a resource template
    which will be creating a VolumeSnapshot.

    At the time that this feature is written, VolumeSnapshots are an Alpha
    feature in Kubernetes. You should check with your Kubernetes Cluster admin
    if they have it enabled.
    """

    def __init__(self,
                 resource_name: str = None,
                 pvc: str = None,
                 snapshot_class: str = None,
                 annotations: Dict[str, str] = None,
                 volume: V1Volume = None,
                 **kwargs):
        """Create a new instance of VolumeSnapshotOp.

        Args:
            resource_name: A desired name for the VolumeSnapshot which will be
                created
            pvc: The name of the PVC which will be snapshotted
            snapshot_class: The snapshot class to use for the dynamically
                created VolumeSnapshot
            annotations: Annotations to be patched in the VolumeSnapshot
            volume: An instance of V1Volume
            kwargs: See ResourceOp definition
        Raises:
            ValueError: if k8s_resource is provided along with other arguments
                        if k8s_resource is not a VolumeSnapshot
                        if pvc and volume are None
                        if pvc and volume are not None
                        if volume does not reference a PVC
        """
        # Add size to output params
        self.attribute_outputs = {"size": "{.status.restoreSize}"}
        # Add default success_condition if None provided
        if "success_condition" not in kwargs:
            kwargs["success_condition"] = "status.readyToUse == true"

        if "k8s_resource" in kwargs:
            if resource_name or pvc or snapshot_class or annotations or volume:
                raise ValueError("You cannot provide k8s_resource along with "
                                 "other arguments.")
            # TODO: Check if is VolumeSnapshot
            super().__init__(**kwargs)
            self.snapshot = V1TypedLocalObjectReference(
                api_group="snapshot.storage.k8s.io",
                kind="VolumeSnapshot",
                name=self.outputs["name"]
            )
            return

        if not (pvc or volume):
            raise ValueError("You must provide a pvc or a volume.")
        elif pvc and volume:
            raise ValueError("You can't provide both pvc and volume.")

        source = None
        deps = []
        if pvc:
            source = V1TypedLocalObjectReference(
                kind="PersistentVolumeClaim",
                name=pvc
            )
        else:
            if not hasattr(volume, "persistent_volume_claim"):
                raise ValueError("The volume must be referencing a PVC.")
            if hasattr(volume, "dependent_names"): #TODO: Replace with type check
                deps = list(volume.dependent_names)
            source = V1TypedLocalObjectReference(
                kind="PersistentVolumeClaim",
                name=volume.persistent_volume_claim.claim_name
            )

        # Set the k8s_resource
        # TODO: Use VolumeSnapshot
        if not match_serialized_pipelineparam(str(resource_name)):
            resource_name = sanitize_k8s_name(resource_name)
        snapshot_metadata = V1ObjectMeta(
            name="{{workflow.name}}-%s" % resource_name,
            annotations=annotations
        )
        k8s_resource = {
            "apiVersion": "snapshot.storage.k8s.io/v1alpha1",
            "kind": "VolumeSnapshot",
            "metadata": snapshot_metadata,
            "spec": {"source": source}
        }
        if snapshot_class:
            k8s_resource["spec"]["snapshotClassName"] = snapshot_class

        super().__init__(
            k8s_resource=k8s_resource,
            **kwargs
        )
        self.dependent_names.extend(deps)
        self.snapshot = V1TypedLocalObjectReference(
            api_group="snapshot.storage.k8s.io",
            kind="VolumeSnapshot",
            name=self.outputs["name"]
        )



# Lint as: python3
# Copyright 2019 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License,Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions for collecting GCP related environment configurations."""

import enum
from typing import List, Text, Optional
from . import utility


class Commands(enum.Enum):
  """Enum for gcloud and gsutil commands."""
  GET_APIS = 1
  GET_CONTAINER_CLUSTERS = 2
  GET_CONTAINER_IMAGES = 3
  GET_DISKS = 4
  GET_GCLOUD_DEFAULT = 5
  GET_NETWORKS = 6
  GET_QUOTAS = 7
  GET_SCOPES = 8
  GET_SERVICE_ACCOUNTS = 9
  GET_STORAGE_BUCKETS = 10
  GET_GCLOUD_VERSION = 11
  GET_AUTH_LIST = 12


_command_string = {
    Commands.GET_APIS: 'services list',
    Commands.GET_CONTAINER_CLUSTERS: 'container clusters list',
    Commands.GET_CONTAINER_IMAGES: 'container images list',
    Commands.GET_DISKS: 'compute disks list',
    Commands.GET_GCLOUD_DEFAULT: 'config list --all',
    Commands.GET_NETWORKS: 'compute networks list',
    Commands.GET_QUOTAS: 'compute regions list',
    Commands.GET_SCOPES: 'compute instances list',
    Commands.GET_SERVICE_ACCOUNTS: 'iam service-accounts list',
    Commands.GET_STORAGE_BUCKETS: 'ls',
    Commands.GET_GCLOUD_VERSION: 'version',
    Commands.GET_AUTH_LIST: 'auth list',
}


def execute_gcloud_command(
    gcloud_command_list: List[Text],
    project_id: Optional[Text] = None,
    human_readable: Optional[bool] = False) -> utility.ExecutorResponse:
  """Function for invoking gcloud command.

  Args:
    gcloud_command_list: a command string list to be past to gcloud example
      format is ['config', 'list', '--all']
    project_id: specificies the project to run the commands against if not
      provided provided will use gcloud default project if one is configured
      otherwise will return an error message.
    human_readable: If false sets parameter --format json for all calls,
      otherwie output will be in human readable format.

  Returns:
    utility.ExecutorResponse with outputs from stdout,stderr and execution code.
  """
  command_list = ['gcloud']
  command_list.extend(gcloud_command_list)
  if not human_readable:
    command_list.extend(['--format', 'json'])

  if project_id is not None:
    command_list.extend(['--project', project_id])

  return utility.ExecutorResponse().execute_command(command_list)


def execute_gsutil_command(
    gsutil_command_list: List[Text],
    project_id: Optional[Text] = None) -> utility.ExecutorResponse:
  """Function for invoking gsutil command.

  This function takes in a gsutil parameter list and returns the results as a
  list of dictionaries.
  Args:
    gsutil_command_list: a command string list to be past to gsutil example
      format is ['config', 'list', '--all']
    project_id: specific project to check the QUOTASs for,if no project id is
      provided will use gcloud default project if one is configured otherwise
      will return an erro massage.

  Returns:
    utility.ExecutorResponse with outputs from stdout,stderr and execution code.
  """
  command_list = ['gsutil']
  command_list.extend(gsutil_command_list)

  if project_id is not None:
    command_list.extend(['-p', project_id])

  return utility.ExecutorResponse().execute_command(command_list)


def get_gcp_configuration(
    configuration: Commands,
    project_id: Optional[Text] = None,
    human_readable: Optional[bool] = False) -> utility.ExecutorResponse:
  """Captures the specified environment configuration.

  Captures the environment configuration for the specified setting such as
  NETWORKSing configuration, project QUOTASs, etc.

  Args:
    configuration: Commands for specific information to be retrieved
      - APIS:  Captures a complete list of enabled APISs and their configuration
        details under the specified project.
      - CONTAINER_CLUSTERS: List all visible k8 clusters under the project.
      - CONTAINER_IMAGES: List of all container images under the project
        container repo.
      - DISKS: List of storage allocated by the project including notebook
        instances as well as k8 pds with corresponding state.
      - GCLOUD_DEFAULT:  Environment default configuration for gcloud
      - NETWORKS: List all NETWORKSs and their configuration under the project.
      - QUOTAS:  Captures a complete list of  QUOTASs for project per
        region,returns the results as a list of dictionaries.
      - SCOPES: list of SCOPESs for each compute resources in the project.
      - SERVICE_ACCOUNTS: List of all service accounts that are enabled under
        this project.
      - STORAGE_BUCKETS: list of buckets and corresponding access information.
    project_id: specific project to check the QUOTASs for,if no project id is
      provided will use gcloud default project if one is configured otherwise
      will return an error message.
    human_readable: If true all output will be in human readable form insted of
      Json.

  Returns:
    A utility.ExecutorResponse with the output results for the specified
    command.
  """
  # storage bucket call requires execute_gsutil_command
  if configuration is Commands.GET_STORAGE_BUCKETS:
    return execute_gsutil_command([_command_string[Commands.GET_STORAGE_BUCKETS]],
                                  project_id)

  # For all other cases can execute the command directly
  return execute_gcloud_command(_command_string[configuration].split(' '),
                                project_id, human_readable)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import inspect
import re
import sys
import tempfile
import logging
import shutil
from collections import OrderedDict
from pathlib import Path
from typing import Callable

from deprecated.sphinx import deprecated

from ..components._components import _create_task_factory_from_component_spec
from ..components._python_op import _func_to_component_spec
from ._container_builder import ContainerBuilder

class VersionedDependency(object):
  """ DependencyVersion specifies the versions """
  def __init__(self, name, version=None, min_version=None, max_version=None):
    """ if version is specified, no need for min_version or max_version;
     if both are specified, version is adopted """
    self._name = name
    if version is not None:
      self._min_version = version
      self._max_version = version
    else:
      self._min_version = min_version
      self._max_version = max_version

  @property
  def name(self):
    return self._name

  @property
  def min_version(self):
    return self._min_version

  @min_version.setter
  def min_version(self, min_version):
    self._min_version = min_version

  def has_min_version(self):
    return self._min_version != None

  @property
  def max_version(self):
    return self._max_version

  @max_version.setter
  def max_version(self, max_version):
    self._max_version = max_version

  def has_max_version(self):
    return self._max_version != None

  def has_versions(self):
    return (self.has_min_version()) or (self.has_max_version())

class DependencyHelper(object):
  """ DependencyHelper manages software dependency information """
  def __init__(self):
    self._PYTHON_PACKAGE = 'PYTHON_PACKAGE'
    self._dependency = {self._PYTHON_PACKAGE:OrderedDict()}

  @property
  def python_packages(self):
    return self._dependency[self._PYTHON_PACKAGE]

  def add_python_package(self, dependency, override=True):
    """ add_single_python_package adds a dependency for the python package

    Args:
      name: package name
      version: it could be a specific version(1.10.0), or a range(>=1.0,<=2.0)
        if not specified, the default is resolved automatically by the pip system.
      override: whether to override the version if already existing in the dependency.
    """
    if dependency.name in self.python_packages and not override:
      return
    self.python_packages[dependency.name] = dependency

  def generate_pip_requirements(self, target_file):
    """ write the python packages to a requirement file
    the generated file follows the order of which the packages are added """
    with open(target_file, 'w') as f:
      for name, version in self.python_packages.items():
        version = self.python_packages[name]
        version_str = ''
        if version.has_min_version():
          version_str += ' >= ' + version.min_version + ','
        if version.has_max_version():
          version_str += ' <= ' + version.max_version + ','
        f.write(name + version_str.rstrip(',') + '\n')

def _dependency_to_requirements(dependency=[], filename='requirements.txt'):
  """
    Generates a requirement file based on the dependency
    Args:
      dependency (list): a list of VersionedDependency, which includes the package name and versions
      filename (str): requirement file name, default as requirements.txt
  """
  dependency_helper = DependencyHelper()
  for version in dependency:
    dependency_helper.add_python_package(version)
  dependency_helper.generate_pip_requirements(filename)

def _generate_dockerfile(filename, base_image, python_version, requirement_filename=None, add_files=None):
  """
    generates dockerfiles
    Args:
      filename (str): target file name for the dockerfile.
      base_image (str): the base image name.
      python_version (str): choose python2 or python3
      requirement_filename (str): requirement file name
      add_files (Dict[str, str]): Map containing the files thats should be added to the container. add_files maps the build context relative source paths to the container destination paths.
  """
  if python_version not in ['python2', 'python3']:
    raise ValueError('python_version has to be either python2 or python3')
  with open(filename, 'w') as f:
    f.write('FROM ' + base_image + '\n')
    if python_version == 'python3':
      f.write('RUN apt-get update -y && apt-get install --no-install-recommends -y -q python3 python3-pip python3-setuptools\n')
    else:
      f.write('RUN apt-get update -y && apt-get install --no-install-recommends -y -q python python-pip python-setuptools\n')
    if requirement_filename is not None:
      f.write('ADD ' + requirement_filename + ' /ml/requirements.txt\n')
      if python_version == 'python3':
        f.write('RUN python3 -m pip install -r /ml/requirements.txt\n')
      else:
        f.write('RUN python -m pip install -r /ml/requirements.txt\n')
    
    for src_path, dst_path in (add_files or {}).items():     
      f.write('ADD ' + src_path + ' ' + dst_path + '\n')


def _configure_logger(logger):
  """ _configure_logger configures the logger such that the info level logs
  go to the stdout and the error(or above) level logs go to the stderr.
  It is important for the Jupyter notebook log rendering """
  if hasattr(_configure_logger, 'configured'):
    # Skip the logger configuration the second time this function
    # is called to avoid multiple streamhandlers bound to the logger.
    return
  setattr(_configure_logger, 'configured', 'true')
  logger.setLevel(logging.INFO)
  info_handler = logging.StreamHandler(stream=sys.stdout)
  info_handler.addFilter(lambda record: record.levelno <= logging.INFO)
  info_handler.setFormatter(logging.Formatter('%(asctime)s:%(levelname)s:%(message)s', datefmt='%Y-%m-%d %H:%M:%S'))
  error_handler = logging.StreamHandler(sys.stderr)
  error_handler.addFilter(lambda record: record.levelno > logging.INFO)
  error_handler.setFormatter(logging.Formatter('%(asctime)s:%(levelname)s:%(message)s', datefmt='%Y-%m-%d %H:%M:%S'))
  logger.addHandler(info_handler)
  logger.addHandler(error_handler)


@deprecated(version='0.1.32', reason='`build_python_component` is deprecated. Use `kfp.containers.build_image_from_working_dir` + `kfp.components.func_to_container_op` instead.')
def build_python_component(component_func, target_image, base_image=None, dependency=[], staging_gcs_path=None, timeout=600, namespace=None, target_component_file=None, python_version='python3'):
  """ build_component automatically builds a container image for the component_func
  based on the base_image and pushes to the target_image.

  Args:
    component_func (python function): The python function to build components upon
    base_image (str): Docker image to use as a base image
    target_image (str): Full URI to push the target image
    staging_gcs_path (str): GCS blob that can store temporary build files
    target_image (str): target image path
    timeout (int): the timeout for the image build(in secs), default is 600 seconds
    namespace (str): the namespace within which to run the kubernetes kaniko job. If the
    job is running on GKE and value is None the underlying functions will use the default namespace from GKE.  .
    dependency (list): a list of VersionedDependency, which includes the package name and versions, default is empty
    python_version (str): choose python2 or python3, default is python3
  Raises:
    ValueError: The function is not decorated with python_component decorator or the python_version is neither python2 nor python3
  """

  _configure_logger(logging.getLogger())

  if component_func is None:
    raise ValueError('component_func must not be None')
  if target_image is None:
    raise ValueError('target_image must not be None')

  if python_version not in ['python2', 'python3']:
    raise ValueError('python_version has to be either python2 or python3')

  if staging_gcs_path is None:
    raise ValueError('staging_gcs_path must not be None')

  if base_image is None:
    base_image = getattr(component_func, '_component_base_image', None)
  if base_image is None:
    from ..components._python_op import default_base_image_or_builder
    base_image = default_base_image_or_builder
    if isinstance(base_image, Callable):
      base_image = base_image()

  logging.info('Build an image that is based on ' +
                                  base_image +
                                  ' and push the image to ' +
                                  target_image)

  component_spec = _func_to_component_spec(component_func, base_image=base_image)
  command_line_args = component_spec.implementation.container.command

  dash_c_index = command_line_args.index('-c')
  program_code_index = dash_c_index + 1
  program_code = command_line_args[program_code_index]
  program_rel_path = 'ml/main.py'
  program_container_path = '/' + program_rel_path

  # Replacing the inline code with calling a local program
  # Before: python3 -u -c 'import sys ...' --param1 ...
  # After:  python3 -u main.py --param1 ...
  command_line_args[program_code_index] = program_container_path
  command_line_args.pop(dash_c_index)

  if python_version == 'python2':
    import warnings
    warnings.warn('Python2 is not longer supported')
    # Replacing the python interpreter
    python_interpreter_index = command_line_args.index('python3')
    command_line_args[python_interpreter_index] = python_version

  arc_docker_filename = 'Dockerfile'
  arc_requirement_filename = 'requirements.txt'

  with tempfile.TemporaryDirectory() as local_build_dir:
    # Write the program code to a file in the context directory
    local_python_filepath = os.path.join(local_build_dir, program_rel_path)
    os.makedirs(os.path.dirname(local_python_filepath), exist_ok=True)
    with open(local_python_filepath, 'w') as f:
      f.write(program_code)

    # Generate the python package requirements file in the context directory
    local_requirement_filepath = os.path.join(local_build_dir, arc_requirement_filename)
    _dependency_to_requirements(dependency, local_requirement_filepath)

    # Generate Dockerfile in the context directory
    local_docker_filepath = os.path.join(local_build_dir, arc_docker_filename)
    _generate_dockerfile(local_docker_filepath, base_image, python_version, arc_requirement_filename, add_files={program_rel_path: program_container_path})

    logging.info('Building and pushing container image.')
    container_builder = ContainerBuilder(staging_gcs_path, target_image, namespace)
    image_name_with_digest = container_builder.build(local_build_dir, arc_docker_filename, target_image, timeout)

  component_spec.implementation.container.image = image_name_with_digest

  # Optionally writing the component definition to a local file for sharing
  target_component_file = target_component_file or getattr(component_func, '_component_target_component_file', None)
  if target_component_file:
    component_spec.save(target_component_file)

  task_factory_function = _create_task_factory_from_component_spec(component_spec)
  return task_factory_function


@deprecated(version='0.1.32', reason='`build_docker_image` is deprecated. Use `kfp.containers.build_image_from_working_dir` instead.')
def build_docker_image(staging_gcs_path, target_image, dockerfile_path, timeout=600, namespace=None):
  """ build_docker_image automatically builds a container image based on the specification in the dockerfile and
  pushes to the target_image.

  Args:
    staging_gcs_path (str): GCS blob that can store temporary build files
    target_image (str): gcr path to push the final image
    dockerfile_path (str): local path to the dockerfile
    timeout (int): the timeout for the image build(in secs), default is 600 seconds
    namespace (str): the namespace within which to run the kubernetes kaniko job. Default is None. If the
    job is running on GKE and value is None the underlying functions will use the default namespace from GKE.  
  """
  _configure_logger(logging.getLogger())

  with tempfile.TemporaryDirectory() as local_build_dir:
    dockerfile_rel_path = 'Dockerfile'
    dst_dockerfile_path = os.path.join(local_build_dir, dockerfile_rel_path)
    shutil.copyfile(dockerfile_path, dst_dockerfile_path)

    container_builder = ContainerBuilder(staging_gcs_path, target_image, namespace=namespace)
    image_name_with_digest = container_builder.build(local_build_dir, dockerfile_rel_path, target_image, timeout)

  logging.info('Build image complete.')
  return image_name_with_digest



# Copyright 2018-2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from . import _container_op
from . import _resource_op
from . import _ops_group
from ._component_bridge import _create_container_op_from_component_and_arguments
from ..components import _components
from ..components._naming import _make_name_unique_by_adding_index
import sys


# This handler is called whenever the @pipeline decorator is applied.
# It can be used by command-line DSL compiler to inject code that runs for every pipeline definition.
_pipeline_decorator_handler = None


def pipeline(name : str = None, description : str = None):
  """Decorator of pipeline functions.

  Usage:
  ```python
  @pipeline(
    name='my awesome pipeline',
    description='Is it really awesome?'
  )
  def my_pipeline(a: PipelineParam, b: PipelineParam):
    ...
  ```
  """
  def _pipeline(func):
    if name:
      func._component_human_name = name
    if description:
      func._component_description = description

    if _pipeline_decorator_handler:
      return _pipeline_decorator_handler(func) or func
    else:
      return func

  return _pipeline

class PipelineConf():
  """PipelineConf contains pipeline level settings
  """
  def __init__(self):
    self.image_pull_secrets = []
    self.timeout = 0
    self.ttl_seconds_after_finished = -1
    self.op_transformers = []
    self.default_pod_node_selector = {}
    self.image_pull_policy = None
    self.parallelism = None

  def set_image_pull_secrets(self, image_pull_secrets):
    """Configures the pipeline level imagepullsecret

    Args:
      image_pull_secrets: a list of Kubernetes V1LocalObjectReference
      For detailed description, check Kubernetes V1LocalObjectReference definition
      https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1LocalObjectReference.md
    """
    self.image_pull_secrets = image_pull_secrets
    return self

  def set_timeout(self, seconds: int):
    """Configures the pipeline level timeout

    Args:
      seconds: number of seconds for timeout
    """
    self.timeout = seconds
    return self

  def set_parallelism(self, max_num_pods: int):
    """Configures the max number of total parallel pods that can execute at the same time in a workflow.

    Args:
        max_num_pods (int): max number of total parallel pods.
    """
    self.parallelism = max_num_pods
    return self

  def set_ttl_seconds_after_finished(self, seconds: int):
    """Configures the ttl after the pipeline has finished.

    Args:
      seconds: number of seconds for the workflow to be garbage collected after it is finished.
    """
    self.ttl_seconds_after_finished = seconds
    return self
  
  def set_default_pod_node_selector(self, label_name: str, value: str): 
    """Add a constraint for nodeSelector for a pipeline. Each constraint is a key-value pair label. For the 
      container to be eligible to run on a node, the node must have each of the constraints appeared
      as labels.

    Args:
        label_name: The name of the constraint label.
        value: The value of the constraint label.
    """
    self.default_pod_node_selector[label_name] = value
    return self
  

  def set_image_pull_policy(self, policy: str):
    """Configures the default image pull policy

    Args:
      policy: the pull policy, has to be one of: Always, Never, IfNotPresent.
      For more info: https://github.com/kubernetes-client/python/blob/10a7f95435c0b94a6d949ba98375f8cc85a70e5a/kubernetes/docs/V1Container.md
    """
    self.image_pull_policy = policy
    return self

  def add_op_transformer(self, transformer):
    """Configures the op_transformers which will be applied to all ops in the pipeline.

    Args:
      transformer: a function that takes a ContainOp as input and returns a ContainerOp
    """
    self.op_transformers.append(transformer)


def get_pipeline_conf():
  """Configure the pipeline level setting to the current pipeline
    Note: call the function inside the user defined pipeline function.
  """
  return Pipeline.get_default_pipeline().conf

#TODO: Pipeline is in fact an opsgroup, refactor the code.
class Pipeline():
  """A pipeline contains a list of operators.

  This class is not supposed to be used by pipeline authors since pipeline authors can use
  pipeline functions (decorated with @pipeline) to reference their pipelines. This class
  is useful for implementing a compiler. For example, the compiler can use the following
  to get the pipeline object and its ops:

  ```python
  with Pipeline() as p:
    pipeline_func(*args_list)

  traverse(p.ops)
  ```
  """

  # _default_pipeline is set when it (usually a compiler) runs "with Pipeline()"
  _default_pipeline = None

  @staticmethod
  def get_default_pipeline():
    """Get default pipeline. """
    return Pipeline._default_pipeline

  @staticmethod
  def add_pipeline(name, description, func):
    """Add a pipeline function with the specified name and description."""
    # Applying the @pipeline decorator to the pipeline function
    func = pipeline(name=name, description=description)(func)

  def __init__(self, name: str):
    """Create a new instance of Pipeline.

    Args:
      name: the name of the pipeline. Once deployed, the name will show up in Pipeline System UI.
    """
    self.name = name
    self.ops = {}
    # Add the root group.
    self.groups = [_ops_group.OpsGroup('pipeline', name=name)]
    self.group_id = 0
    self.conf = PipelineConf()
    self._metadata = None

  def __enter__(self):
    if Pipeline._default_pipeline:
      raise Exception('Nested pipelines are not allowed.')

    Pipeline._default_pipeline = self
    self._old_container_task_constructor = _components._container_task_constructor
    _components._container_task_constructor = _create_container_op_from_component_and_arguments

    def register_op_and_generate_id(op):
      return self.add_op(op, op.is_exit_handler)

    self._old__register_op_handler = _container_op._register_op_handler
    _container_op._register_op_handler = register_op_and_generate_id
    return self

  def __exit__(self, *args):
    Pipeline._default_pipeline = None
    _container_op._register_op_handler = self._old__register_op_handler
    _components._container_task_constructor = self._old_container_task_constructor

  def add_op(self, op: _container_op.BaseOp, define_only: bool):
    """Add a new operator.

    Args:
      op: An operator of ContainerOp, ResourceOp or their inherited types.

    Returns
      op_name: a unique op name.
    """
    #If there is an existing op with this name then generate a new name.
    op_name = _make_name_unique_by_adding_index(op.human_name, list(self.ops.keys()), ' ')

    self.ops[op_name] = op
    if not define_only:
      self.groups[-1].ops.append(op)

    return op_name

  def push_ops_group(self, group: _ops_group.OpsGroup):
    """Push an OpsGroup into the stack.

    Args:
      group: An OpsGroup. Typically it is one of ExitHandler, Branch, and Loop.
    """
    self.groups[-1].groups.append(group)
    self.groups.append(group)

  def pop_ops_group(self):
    """Remove the current OpsGroup from the stack."""
    del self.groups[-1]

  def remove_op_from_groups(self, op):
    for group in self.groups:
      group.remove_op_recursive(op)

  def get_next_group_id(self):
    """Get next id for a new group. """

    self.group_id += 1
    return self.group_id

  def _set_metadata(self, metadata):
    '''_set_metadata passes the containerop the metadata information
    Args:
      metadata (ComponentMeta): component metadata
    '''
    self._metadata = metadata



# Lint as: python3
# Copyright 2019 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License,Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Supporting tools and classes for diagnose_me."""

import json
import subprocess
from typing import List, Text


class ExecutorResponse(object):
  """Class for keeping track of output of _executor methods.

  Data model for executing commands and capturing their response. This class
  defines the data model layer for execution results, based on MVC design
  pattern.

  TODO() This class should be extended to contain data structure to better
  represent the underlying data instaed of dict for various response types.
  """

  def execute_command(self, command_list: List[Text]):
    """Executes the command in command_list.

    sets values for _stdout,_std_err, and returncode accordingly.

    TODO(): This method is kept in ExecutorResponse for simplicity, however this
    deviates from MVP design pattern. It should be factored out in future.

    Args:
      command_list: A List of strings that represts the command and parameters
        to be executed.

    Returns:
      Instance of utility.ExecutorResponse.
    """

    try:
      # TODO() switch to process.run to simplify the code.
      process = subprocess.Popen(
          command_list, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
      stdout, stderr = process.communicate()
      self._stdout = stdout.decode('utf-8')
      self._stderr = stderr.decode('utf-8')
      self._returncode = process.returncode
    except OSError as e:
      self._stderr = e
      self._stdout = ''
      self._returncode = e.errno
    self._parse_raw_input()
    return self

  def _parse_raw_input(self):
    """Parses the raw input and popluates _json and _parsed properies."""
    try:
      self._parsed_output = json.loads(self._stdout)
      self._json = self._stdout
    except json.JSONDecodeError:
      self._json = json.dumps(self._stdout)
      self._parsed_output = self._stdout

  @property
  def parsed_output(self) -> Text:
    """Json load results of stdout or raw results if stdout was not Json."""
    return self._parsed_output

  @property
  def has_error(self) -> bool:
    """Returns true if execution error code was not 0."""
    return self._returncode != 0

  @property
  def json_output(self) -> Text:
    """Run results in stdout in json format."""
    return self._parsed_output

  @property
  def stderr(self):
    return self._stderr



import re
from typing import List, Union, Dict, Text, Any, Tuple, Optional

from kfp import dsl

ItemList = List[Union[int, float, str, Dict[Text, Any]]]


class LoopArguments(dsl.PipelineParam):
    """Class representing the arguments that are looped over in a ParallelFor loop in the KFP DSL.
    This doesn't need to be instantiated by the end user, rather it will be automatically created by a
    ParallelFor ops group."""
    LOOP_ITEM_NAME_BASE = 'loop-item'
    LOOP_ITEM_PARAM_NAME_BASE = 'loop-item-param'
    # number of characters in the code which is passed to the constructor
    NUM_CODE_CHARS = 8
    LEGAL_SUBVAR_NAME_REGEX = re.compile(r'[a-zA-Z_][0-9a-zA-Z_]*')

    @classmethod
    def _subvar_name_is_legal(cls, proposed_variable_name: Text):
        return re.match(cls.LEGAL_SUBVAR_NAME_REGEX, proposed_variable_name) is not None

    def __init__(self, items: Union[ItemList, dsl.PipelineParam], code: Text, name_override: Optional[Text]=None, op_name: Optional[Text]=None, *args, **kwargs):
        """LoopArguments represent the set of items to loop over in a ParallelFor loop.  This class shouldn't be
        instantiated by the user but rather is created by _ops_group.ParallelFor.

        Args:
            items: List of items to loop over.  If a list of dicts then, all dicts must have the same keys and every
                key must be a legal Python variable name.
            code: A unique code used to identify these loop arguments.  Should match the code for the ParallelFor
                ops_group which created these _LoopArguments.  This prevents parameter name collisions.
        """
        if name_override is None:
            super().__init__(name=self._make_name(code), *args, **kwargs)
        else:
            super().__init__(name=name_override, op_name=op_name, *args, **kwargs)

        if not isinstance(items, (list, tuple, dsl.PipelineParam)):
            raise TypeError("Expected list, tuple, or PipelineParam, got {}.".format(type(items)))

        if isinstance(items, tuple):
            items = list(items)

        if isinstance(items, list) and isinstance(items[0], dict):
            subvar_names = set(items[0].keys())
            for item in items:
                if not set(item.keys()) == subvar_names:
                    raise ValueError("If you input a list of dicts then all dicts should have the same keys. "
                                     "Got: {}.".format(items))

            # then this block creates loop_args.variable_a and loop_args.variable_b
            for subvar_name in subvar_names:
                if not self._subvar_name_is_legal(subvar_name):
                    raise ValueError("Tried to create subvariable named {} but that's not a legal Python variable "
                                     "name.".format(subvar_name))
                setattr(self, subvar_name, LoopArgumentVariable(self.name, subvar_name, loop_args_op_name=self.op_name))

        self.items_or_pipeline_param = items
        self.referenced_subvar_names = []

    @classmethod
    def from_pipeline_param(cls, param: dsl.PipelineParam) -> 'LoopArguments':
        return LoopArguments(
            items=param,
            code=None,
            name_override=param.name + '-' + cls.LOOP_ITEM_NAME_BASE,
            op_name=param.op_name,
            value=param.value,
        )

    def __getattr__(self, item):
        # this is being overridden so that we can access subvariables of the LoopArguments (i.e.: item.a) without
        # knowing the subvariable names ahead of time
        self.referenced_subvar_names.append(item)
        return LoopArgumentVariable(self.name, item, loop_args_op_name=self.op_name)

    def to_list_for_task_yaml(self):
        if isinstance(self.items_or_pipeline_param, (list, tuple)):
            return self.items_or_pipeline_param
        else:
            raise ValueError("You should only call this method on loop args which have list items, "
                             "not pipeline param items.")

    @classmethod
    def _make_name(cls, code: Text):
        """Make a name for this parameter.  Code is a """
        return '{}-{}'.format(cls.LOOP_ITEM_PARAM_NAME_BASE, code)

    @classmethod
    def name_is_withitems_loop_argument(cls, param_name: Text) -> bool:
        """Return True if the given parameter name looks like it came from a loop arguments parameter."""
        return re.match(
            '%s-[0-9a-f]{%s}' % (cls.LOOP_ITEM_PARAM_NAME_BASE, cls.NUM_CODE_CHARS),
            param_name,
        ) is not None

    @classmethod
    def name_is_withparams_loop_argument(cls, param_name: Text) -> bool:
        """Return True if the given parameter name looks like it came from a withParams loop item."""
        return ('-' + cls.LOOP_ITEM_NAME_BASE) in param_name

class LoopArgumentVariable(dsl.PipelineParam):
    """Represents a subvariable for loop arguments.  This is used for cases where we're looping over maps,
    each of which contains several variables."""
    SUBVAR_NAME_DELIMITER = '-subvar-'

    def __init__(
        self,
        loop_args_name: Text,
        this_variable_name: Text,
        loop_args_op_name: Text,
    ):
        """
        If the user ran:
            with dsl.ParallelFor([{'a': 1, 'b': 2}, {'a': 3, 'b': 4}]) as item:
                ...
        Then there's be one _LoopArgumentsVariable for 'a' and another for 'b'.

        Args:
            loop_args_name:  the name of the _LoopArguments object that this is a subvariable to.
            this_variable_name: the name of this subvariable, which is the name of the dict key that spawned
                this subvariable.
        """
        super().__init__(
            name=self.get_name(loop_args_name=loop_args_name,
            this_variable_name=this_variable_name),
            op_name=loop_args_op_name,
        )

    @classmethod
    def get_name(cls, loop_args_name: Text, this_variable_name: Text) -> Text:
        """Get the name

        Args:
            loop_args_name: the name of the loop args parameter that this LoopArgsVariable is attached to.
            this_variable_name: the name of this LoopArgumentsVariable subvar.

        Returns: The name of this loop args variable.
        """
        return '{}{}{}'.format(loop_args_name, cls.SUBVAR_NAME_DELIMITER, this_variable_name)

    @classmethod
    def name_is_loop_arguments_variable(cls, param_name: Text) -> bool:
        """Return True if the given parameter name looks like it came from a LoopArgumentsVariable."""
        return re.match(
            '.+%s.+' % cls.SUBVAR_NAME_DELIMITER,
            param_name
        ) is not None

    @classmethod
    def parse_loop_args_name_and_this_var_name(cls, t: Text) -> Tuple[Text, Text]:
        """Get the loop arguments param name and this subvariable name from the given parameter name."""
        m = re.match('(?P<loop_args_name>.*){}(?P<this_var_name>.*)'.format(cls.SUBVAR_NAME_DELIMITER), t)
        if m is None:
            return None
        else:
            return m.groupdict()['loop_args_name'], m.groupdict()['this_var_name']

    @classmethod
    def get_subvar_name(cls, t: Text) -> Text:
        """Get the subvariable name from a given LoopArgumentsVariable parameter name."""
        out = cls.parse_loop_args_name_and_this_var_name(t)
        if out is None:
            raise ValueError("Couldn't parse variable name: {}".format(t))
        return out[1]



__all__ = [
    'Container',
    'PodArgoSubset',
]


from collections import OrderedDict

from typing import Any, Dict, List, Mapping, Optional, Sequence, Union

from ...modelbase import ModelBase


class EnvVar(ModelBase):
    _serialized_names = {
        'value_from': 'valueFrom',
    }
    def __init__(self,
        name: str,
        value: Optional[str] = None,
        #value_from: Optional[EnvVarSource] = None, #TODO: Add if needed
    ):
        super().__init__(locals())


class ExecAction(ModelBase):
    def __init__(self,
        command: List[str],
    ):
        super().__init__(locals())


class Handler(ModelBase):
    _serialized_names = {
        'http_get': 'httpGet',
        'tcp_socket': 'tcpSocket',
    }
    def __init__(self,
        exec: Optional[ExecAction] = None,
        #http_get: Optional[HTTPGetAction] = None, #TODO: Add if needed
        #tcp_socket: Optional[TCPSocketAction] = None, #TODO: Add if needed
    ):
        super().__init__(locals())


class Lifecycle(ModelBase):
    _serialized_names = {
        'post_start': 'postStart',
        'pre_stop': 'preStop',
    }
    def __init__(self,
        post_start: Optional[Handler] = None,
        pre_stop: Optional[Handler] = None,
    ):
        super().__init__(locals())


class VolumeMount(ModelBase):
    _serialized_names = {
        'mount_path': 'mountPath',
        'mount_propagation': 'mountPropagation',
        'read_only': 'readOnly',
        'sub_path': 'subPath',
    }
    def __init__(self,
        name: str,
        mount_path: str,
        mount_propagation: Optional[str] = None,
        read_only: Optional[bool] = None,
        sub_path: Optional[str] = None,
    ):
        super().__init__(locals())


class ResourceRequirements(ModelBase):
    def __init__(self,
        limits: Optional[Dict[str, str]] = None,
        requests: Optional[Dict[str, str]] = None,
    ):
        super().__init__(locals())


class ContainerPort(ModelBase):
    _serialized_names = {
        'container_port': 'containerPort',
        'host_ip': 'hostIP',
        'host_port': 'hostPort',
    }
    def __init__(self,
        container_port: int,
        host_ip: Optional[str] = None,
        host_port: Optional[int] = None,
        name: Optional[str] = None,
        protocol: Optional[str] = None,
    ):
        super().__init__(locals())


class VolumeDevice(ModelBase):
    _serialized_names = {
        'device_path': 'devicePath',
    }
    def __init__(self,
        device_path: str,
        name: str,
    ):
        super().__init__(locals())


class Probe(ModelBase):
    _serialized_names = {
        'failure_threshold': 'failureThreshold',
        'http_get': 'httpGet',
        'initial_delay_seconds': 'initialDelaySeconds',
        'period_seconds': 'periodSeconds',
        'success_threshold': 'successThreshold',
        'tcp_socket': 'tcpSocket',
        'timeout_seconds': 'timeoutSeconds'
    }
    def __init__(self,
        exec: Optional[ExecAction] = None,
        failure_threshold: Optional[int] = None,
        #http_get: Optional[HTTPGetAction] = None, #TODO: Add if needed
        initial_delay_seconds: Optional[int] = None,
        period_seconds: Optional[int] = None,
        success_threshold: Optional[int] = None,
        #tcp_socket: Optional[TCPSocketAction] = None, #TODO: Add if needed
        timeout_seconds: Optional[int] = None,
    ):
        super().__init__(locals())


class SecurityContext(ModelBase):
    _serialized_names = {
        'allow_privilege_escalation': 'allowPrivilegeEscalation',
        'capabilities': 'capabilities',
        'privileged': 'privileged',
        'read_only_root_filesystem': 'readOnlyRootFilesystem',
        'run_as_group': 'runAsGroup',
        'run_as_non_root': 'runAsNonRoot',
        'run_as_user': 'runAsUser',
        'se_linux_options': 'seLinuxOptions'
    }
    def __init__(self,
        allow_privilege_escalation: Optional[bool] = None,
        #capabilities: Optional[Capabilities] = None, #TODO: Add if needed
        privileged: Optional[bool] = None,
        read_only_root_filesystem: Optional[bool] = None,
        run_as_group: Optional[int] = None,
        run_as_non_root: Optional[bool] = None,
        run_as_user: Optional[int] = None,
        #se_linux_options: Optional[SELinuxOptions] = None, #TODO: Add if needed
    ):
        super().__init__(locals())


class Container(ModelBase):
    _serialized_names = {
        'env_from': 'envFrom',
        'image_pull_policy': 'imagePullPolicy',
        'liveness_probe': 'livenessProbe',
        'readiness_probe': 'readinessProbe',
        'security_context': 'securityContext',
        'stdin_once': 'stdinOnce',
        'termination_message_path': 'terminationMessagePath',
        'termination_message_policy': 'terminationMessagePolicy',
        'volume_devices': 'volumeDevices',
        'volume_mounts': 'volumeMounts',
        'working_dir': 'workingDir',
    }

    def __init__(self,
        #Better to set at Component level
        image: Optional[str] = None,
        command: Optional[List[str]] = None,
        args: Optional[List[str]] = None,
        env: Optional[List[EnvVar]] = None,

        working_dir: Optional[str] = None, #Not really needed: container should have proper working dir set up

        lifecycle: Optional[Lifecycle] = None, #Can be used to specify pre-exit commands to run TODO: Probably support at Component level.

        #Better to set at Task level
        volume_mounts: Optional[List[VolumeMount]] = None,
        resources: Optional[ResourceRequirements] = None,

        #Might not be used a lot
        ports: Optional[List[ContainerPort]] = None,
        #env_from: Optional[List[EnvFromSource]] = None, #TODO: Add if needed
        volume_devices: Optional[List[VolumeDevice]] = None,

        #Probably not needed
        name: Optional[str] = None, #Required by k8s schema, but not Argo.
        image_pull_policy: Optional[str] = None,
        liveness_probe: Optional[Probe] = None,
        readiness_probe: Optional[Probe] = None,
        security_context: Optional[SecurityContext] = None,
        stdin: Optional[bool] = None,
        stdin_once: Optional[bool] = None,
        termination_message_path: Optional[str] = None,
        termination_message_policy: Optional[str] = None,
        tty: Optional[bool] = None,
    ):
        super().__init__(locals())


#class NodeAffinity(ModelBase):
#    _serialized_names = {
#        'preferred_during_scheduling_ignored_during_execution': 'preferredDuringSchedulingIgnoredDuringExecution',
#        'required_during_scheduling_ignored_during_execution': 'requiredDuringSchedulingIgnoredDuringExecution',
#    }
#    def __init__(self,
#        preferred_during_scheduling_ignored_during_execution: Optional[List[PreferredSchedulingTerm]] = None,
#        required_during_scheduling_ignored_during_execution: Optional[NodeSelector] = None,
#    ):
#        super().__init__(locals())


#class Affinity(ModelBase):
#    _serialized_names = {
#        'node_affinity': 'nodeAffinity',
#        'pod_affinity': 'podAffinity',
#        'pod_anti_affinity': 'podAntiAffinity',
#    }
#    def __init__(self,
#        node_affinity: Optional[NodeAffinity] = None,
#        #pod_affinity: Optional[PodAffinity] = None, #TODO: Add if needed
#        #pod_anti_affinity: Optional[PodAntiAffinity] = None, #TODO: Add if needed
#    ):
#        super().__init__(locals())


class Toleration(ModelBase):
    _serialized_names = {
        'toleration_seconds': 'tolerationSeconds',
    }
    def __init__(self,
        effect: Optional[str] = None,
        key: Optional[str] = None,
        operator: Optional[str] = None,
        toleration_seconds: Optional[int] = None,
        value: Optional[str] = None,
    ):
        super().__init__(locals())


class KeyToPath(ModelBase):
    def __init__(self,
        key: str,
        path: str,
        mode: Optional[int] = None,
    ):
        super().__init__(locals())


class SecretVolumeSource(ModelBase):
    _serialized_names = {
        'default_mode': 'defaultMode',
        'secret_name': 'secretName'
    }
    def __init__(self,
        default_mode: Optional[int] = None,
        items: Optional[List[KeyToPath]] = None,
        optional: Optional[bool] = None,
        secret_name: Optional[str] = None,
    ):
        super().__init__(locals())


class NFSVolumeSource(ModelBase):
    _serialized_names = {
        'read_only': 'readOnly',
    }
    def __init__(self,
        path: str,
        server: str,
        read_only: Optional[bool] = None,
    ):
        super().__init__(locals())


class PersistentVolumeClaimVolumeSource(ModelBase):
    _serialized_names = {
        'claim_name': 'claimName',
        'read_only': 'readOnly'
    }
    def __init__(self,
        claim_name: str,
        read_only: Optional[bool] = None,
    ):
        super().__init__(locals())


class Volume(ModelBase):
    _serialized_names = {
        'aws_elastic_block_store': 'awsElasticBlockStore',
        'azure_disk': 'azureDisk',
        'azure_file': 'azureFile',
        'cephfs': 'cephfs',
        'cinder': 'cinder',
        'config_map': 'configMap',
        'downward_api': 'downwardAPI',
        'empty_dir': 'emptyDir',
        'fc': 'fc',
        'flex_volume': 'flexVolume',
        'flocker': 'flocker',
        'gce_persistent_disk': 'gcePersistentDisk',
        'git_repo': 'gitRepo',
        'glusterfs': 'glusterfs',
        'host_path': 'hostPath',
        'iscsi': 'iscsi',
        'name': 'name',
        'nfs': 'nfs',
        'persistent_volume_claim': 'persistentVolumeClaim',
        'photon_persistent_disk': 'photonPersistentDisk',
        'portworx_volume': 'portworxVolume',
        'projected': 'projected',
        'quobyte': 'quobyte',
        'rbd': 'rbd',
        'scale_io': 'scaleIO',
        'secret': 'secret',
        'storageos': 'storageos',
        'vsphere_volume': 'vsphereVolume'
    }

    def __init__(self,
        name: str,
        secret: Optional[SecretVolumeSource] = None,
        nfs: Optional[NFSVolumeSource] = None,
        persistent_volume_claim: Optional[PersistentVolumeClaimVolumeSource] = None,

        #No validation for these volume types
        aws_elastic_block_store: Optional[Mapping] = None, #AWSElasticBlockStoreVolumeSource,
        azure_disk: Optional[Mapping] = None, #AzureDiskVolumeSource,
        azure_file: Optional[Mapping] = None, #AzureFileVolumeSource,
        cephfs: Optional[Mapping] = None, #CephFSVolumeSource,
        cinder: Optional[Mapping] = None, #CinderVolumeSource,
        config_map: Optional[Mapping] = None, #ConfigMapVolumeSource,
        downward_api: Optional[Mapping] = None, #DownwardAPIVolumeSource,
        empty_dir: Optional[Mapping] = None, #EmptyDirVolumeSource,
        fc: Optional[Mapping] = None, #FCVolumeSource,
        flex_volume: Optional[Mapping] = None, #FlexVolumeSource,
        flocker: Optional[Mapping] = None, #FlockerVolumeSource,
        gce_persistent_disk: Optional[Mapping] = None, #GCEPersistentDiskVolumeSource,
        git_repo: Optional[Mapping] = None, #GitRepoVolumeSource,
        glusterfs: Optional[Mapping] = None, #GlusterfsVolumeSource,
        host_path: Optional[Mapping] = None, #HostPathVolumeSource,
        iscsi: Optional[Mapping] = None, #ISCSIVolumeSource,
        photon_persistent_disk: Optional[Mapping] = None, #PhotonPersistentDiskVolumeSource,
        portworx_volume: Optional[Mapping] = None, #PortworxVolumeSource,
        projected: Optional[Mapping] = None, #ProjectedVolumeSource,
        quobyte: Optional[Mapping] = None, #QuobyteVolumeSource,
        rbd: Optional[Mapping] = None, #RBDVolumeSource,
        scale_io: Optional[Mapping] = None, #ScaleIOVolumeSource,
        storageos: Optional[Mapping] = None, #StorageOSVolumeSource,
        vsphere_volume: Optional[Mapping] = None, #VsphereVirtualDiskVolumeSource,
    ):
        super().__init__(locals())


class PodSpecArgoSubset(ModelBase):
    _serialized_names = {
        'active_deadline_seconds': 'activeDeadlineSeconds',
        'affinity': 'affinity',
        #'automount_service_account_token': 'automountServiceAccountToken',
        #'containers': 'containers',
        #'dns_config': 'dnsConfig',
        #'dns_policy': 'dnsPolicy',
        #'host_aliases': 'hostAliases',
        #'host_ipc': 'hostIPC',
        #'host_network': 'hostNetwork',
        #'host_pid': 'hostPID',
        #'hostname': 'hostname',
        #'image_pull_secrets': 'imagePullSecrets',
        #'init_containers': 'initContainers',
        #'node_name': 'nodeName',
        'node_selector': 'nodeSelector',
        #'priority': 'priority',
        #'priority_class_name': 'priorityClassName',
        #'readiness_gates': 'readinessGates',
        #'restart_policy': 'restartPolicy',
        #'scheduler_name': 'schedulerName',
        #'security_context': 'securityContext',
        #'service_account': 'serviceAccount',
        #'service_account_name': 'serviceAccountName',
        #'share_process_namespace': 'shareProcessNamespace',
        #'subdomain': 'subdomain',
        #'termination_grace_period_seconds': 'terminationGracePeriodSeconds',
        'tolerations': 'tolerations',
        'volumes': 'volumes',
    }
    def __init__(self,
        active_deadline_seconds: Optional[int] = None,
        affinity: Optional[Mapping] = None, #Affinity, #No validation
        #automount_service_account_token: Optional[bool] = None, #Not supported by Argo
        #containers: Optional[List[Container]] = None, #Not supported by Argo
        #dns_config: Optional[PodDNSConfig] = None, #Not supported by Argo
        #dns_policy: Optional[str] = None, #Not supported by Argo
        #host_aliases: Optional[List[HostAlias]] = None, #Not supported by Argo
        #host_ipc: Optional[bool] = None, #Not supported by Argo
        #host_network: Optional[bool] = None, #Not supported by Argo
        #host_pid: Optional[bool] = None, #Not supported by Argo
        #hostname: Optional[str] = None, #Not supported by Argo
        #image_pull_secrets: Optional[List[LocalObjectReference]] = None, #Not supported by Argo
        #init_containers: Optional[List[Container]] = None, #Not supported by Argo
        #node_name: Optional[str] = None, #Not supported by Argo
        node_selector: Optional[Dict[str, str]] = None,
        #priority: Optional[int] = None, #Not supported by Argo
        #priority_class_name: Optional[str] = None, #Not supported by Argo
        #readiness_gates: Optional[List[PodReadinessGate]] = None, #Not supported by Argo
        #restart_policy: Optional[str] = None, #Not supported by Argo
        #scheduler_name: Optional[str] = None, #Not supported by Argo
        #security_context: Optional[PodSecurityContext] = None, #Not supported by Argo
        #service_account: Optional[str] = None, #Not supported by Argo
        #service_account_name: Optional[str] = None, #Not supported by Argo
        #share_process_namespace: Optional[bool] = None, #Not supported by Argo
        #subdomain: Optional[str] = None, #Not supported by Argo
        #termination_grace_period_seconds: Optional[int] = None, #Not supported by Argo
        tolerations: Optional[List[Toleration]] = None,
        volumes: Optional[List[Volume]] = None, #Argo only supports volumes at the Workflow level

        #+Argo features:
        #+Metadata: ArgoMetadata? (Argo version)
        #+RetryStrategy: ArgoRetryStrategy ~= k8s.JobSpec.backoffLimit
        #+Parallelism: int
    ):
        super().__init__(locals())


class ObjectMetaArgoSubset(ModelBase):
    def __init__(self,
        annotations: Optional[Dict[str, str]] = None,
        labels: Optional[Dict[str, str]] = None,
    ):
        super().__init__(locals())


class PodArgoSubset(ModelBase):
    _serialized_names = {
        'api_version': 'apiVersion',
        'kind': 'kind',
        'metadata': 'metadata',
        'spec': 'spec',
        'status': 'status',
    }
    def __init__(self,
        #api_version: Optional[str] = None,
        #kind: Optional[str] = None,
        #metadata: Optional[ObjectMeta] = None,
        metadata: Optional[ObjectMetaArgoSubset] = None,
        #spec: Optional[PodSpec] = None,
        spec: Optional[PodSpecArgoSubset] = None,
        #status: Optional[PodStatus] = None,
    ):
        super().__init__(locals())



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from .types import BaseType, _check_valid_type_dict
from ..components._data_passing import serialize_value
from ..components.structures import ComponentSpec, InputSpec, OutputSpec


def _annotation_to_typemeta(annotation):
  '''_annotation_to_type_meta converts an annotation to a type structure
  Args:
    annotation(BaseType/str/dict): input/output annotations
      BaseType: registered in kfp.dsl.types
      str: either a string of a dict serialization or a string of the type name
      dict: type name and properties. note that the properties values can be dict.
  Returns:
    dict or string representing the type
    '''
  if isinstance(annotation, BaseType):
    arg_type = annotation.to_dict()
  elif isinstance(annotation, str):
    arg_type = annotation
  elif isinstance(annotation, dict):
    if not _check_valid_type_dict(annotation):
      raise ValueError('Annotation ' + str(annotation) + ' is not a valid type dictionary.')
    arg_type = annotation
  else:
    return None
  return arg_type


def _extract_pipeline_metadata(func):
  '''Creates pipeline metadata structure instance based on the function signature.'''

  # Most of this code is only needed for verifying the default values against "openapi_schema_validator" type properties.
  # TODO: Move the value verification code to some other place

  from ._pipeline_param import PipelineParam

  import inspect
  fullargspec = inspect.getfullargspec(func)
  args = fullargspec.args
  annotations = fullargspec.annotations

  # defaults
  arg_defaults = {}
  if fullargspec.defaults:
    for arg, default in zip(reversed(fullargspec.args), reversed(fullargspec.defaults)):
      arg_defaults[arg] = default

  for arg in args:
    arg_type = None
    arg_default = arg_defaults[arg] if arg in arg_defaults else None
    if isinstance(arg_default, PipelineParam):
      warnings.warn('Explicit creation of `kfp.dsl.PipelineParam`s by the users is deprecated. The users should define the parameter type and default values using standard pythonic constructs: def my_func(a: int = 1, b: str = "default"):')
      arg_default = arg_default.value
    if arg in annotations:
      arg_type = _annotation_to_typemeta(annotations[arg])
    arg_type_properties = list(arg_type.values())[0] if isinstance(arg_type, dict) else {}
    if 'openapi_schema_validator' in arg_type_properties and arg_default is not None:
      from jsonschema import validate
      import json
      schema_object = arg_type_properties['openapi_schema_validator']
      if isinstance(schema_object, str):
        # In case the property value for the schema validator is a string instead of a dict.
        schema_object = json.loads(schema_object)
      # Only validating non-serialized values
      validate(instance=arg_default, schema=schema_object)



  from kfp.components._python_op import _extract_component_interface
  component_spec = _extract_component_interface(func)
  return component_spec
